/*----------------------------------------------------------------------------*/
/* approyrd.c - application dependent routines, version of appxxxxx.c changed */
/*              to prepare for solution of each particular problem.           */
/*                                                                            */
/* This application, approyrd.c, is an implementation of the particular Royal */
/* Road function defined by Prof. John Holland, at the 1993 Fifth Internat.'l */
/* Conference on Genetic Algorithms and their Applications, Urbana-Champaign, */
/* Illinois.  He posed it as a challenge problem to test the performance of   */
/* genetic algorithms.  He indicated that, with the parameter settings of     */
/* SCHEMA = 8, INTRON = 7, COE_UU = 1.0, COE_U = 0.3, COE_V = 0.02, and       */
/* M_LTD = 4, he could attain royal_road_level 3 most of the time within      */
/* 10,000 function evaluations.  He challenged other GA users to match or beat*/
/* that performance.  He indicated that he used a population size of 512 to   */
/* obtain his solutions, and did NOT use a "simple genetic algorithm."        */
/*                                                                            */
/* The chromosome is structured as: (SCHEMA bits, INTRON bits) ** 2k, for     */
/* some k = 1,2,3, or 4.  If k=1, lchrom = 30; if k=2, lchrom = 60; if k = 3, */
/* lchrom = 120, and if k = 4 (as Holland used it), lchrom = 240.  Thus, the  */
/* challenge Holland presented was to get either the first 8 groups all 1, or */
/* the last 8 groups all 1 (== royal road level 3), on a chromosome with 16   */
/* total groups (length 240).                                                 */
/*                                                                            */
/* UNLESS YOU CHANGE THE DEFINES BELOW, YOU SHOULD RUN THIS PROBLEM WITH      */
/* alpha_size of 2 (binary), and numfields as one of: 60, 120, OR 240.        */
/*                                                                            */
/* In this example, the utility field pointer of each chromosome is used to   */
/* maintain royal road variables (see struct rr_type).                        */
/*----------------------------------------------------------------------------*/

#include "external.h"


#define SCHEMA  8       /* length of target schemata for RR */
#define INTRON  7       /* # bits between target schemata for RR */
#define COE_UU  1.0     /* base bonus for RR fn  -- Holland's U* */
#define COE_U   0.3     /* increment for lowest lev match, u     */
#define COE_V   0.02    /* incr or decr amt , Holland's v        */
#define M_LTD   4       /* Holland's m*, # bits in lowest level  */
			/* schemata after which start to accrue  */
			/* a penalty, not a bonus, for more 1's  */

static int max_royal_road_level;    /* used to hold size of rr problem being
				 * solved (max value royal_road_level could 
				 * possibly attain) */

static int print_rr_strings = 0;        /* flag for special rr-format print  */
struct rr_type {        /* for contents of utility fields */
    int     royal_road_level;
    float   bonus;
    float   part;
};
static int savedrr_level[64];   /* for trigger of quiet-mode output */
static int first_time_stats = 1;
static int group_rec[16];


void     writechrom_rr();
void     writepop_rr();

void
objfunc(critter)
/* Application dependent objective function. THIS IS THE ONLY ROUTINE THE */
/* USER ABSOLUTELY MUST FILL IN TO DEFINE THE USER'S PROBLEM.  OTHERS ARE */
/* OPTIONAL, DEPENDING ON THE NATURE AND COMPLEXITY OF THE PROBLEM.       */
struct individual *critter;

/*      This is an objective function for computing the fitness_value for
	Holland's 1993 ICGA version of the Royal Road Problem.
	It has been corrected per GAList vol 7 no. 23, 8/26/93, by EDG, on
	10/12/93.  Now, no bonus points are awarded for a given level until
	it has been achieved (fixing Holland's coding error in GAList) */
{
    int     n[32], m[32];
    int     ii, jj, s, pos;
    int     block, offset;
    unsigned tp, mask = 1;
    float   bonus, part;

    int     power();

    neval++;
    local_cycle_neval++;
    critter->neval = neval;

    bonus = 0.0;
    ((struct rr_type *) (critter->utility))->royal_road_level = 0;

    for (ii = 0; ii < max_royal_road_level + 1; ii++) {
	s = power(max_royal_road_level - ii);
	n[ii] = 0;

	for (jj = 0; jj < s; jj++) {
	    if (rr_func(critter, ii, jj) == 1)
		n[ii]++;

	}
	if (n[ii] != 0) {
	    bonus += COE_UU + (n[ii] - 1) * COE_U;
	    ((struct rr_type *) (critter->utility))->royal_road_level = ii;
	}
    }

    s = power(max_royal_road_level);
    part = 0.0;

    for (ii = 0; ii < s; ii++) {
	m[ii] = 0;

	for (jj = 0; jj < SCHEMA; jj++) {
	    pos = jj + ii * (SCHEMA + INTRON);
	    block = pos / UINTSIZE;
	    offset = pos % UINTSIZE;
	    tp = critter->chrom[block];
	    if (offset > 0)
		tp = tp >> offset;
	    if ((tp & mask) == 1)
		m[ii]++;
	}
	if (m[ii] < M_LTD + 1)
	    part += m[ii] * COE_V;
	if ((m[ii] > M_LTD) && (m[ii] < SCHEMA))
	    part += -(m[ii] - M_LTD) * COE_V;
	if (m[ii] == SCHEMA)
	    part += 0.0;
    }
    ((struct rr_type *) (critter->utility))->bonus = bonus;
    ((struct rr_type *) (critter->utility))->part = part;
    critter->init_fitness = bonus + part;
}


int
power(n)
int     n;
{
    int     i, temp;

    if (n == 0)
	return (1);
    else {
	temp = 1;
	for (i = 0; i < n; i++)
	    temp = temp * 2;
	return (temp);
    }
}


int
rr_func(critter, i, j)
struct individual *critter;
int     i, j;

/*      This function is used to identify if the jth schema at level i exists.
	     rr_func = 1;    if the jth target schema at level i exists.
	     rr_func = 0;    if the jth target schema at level i does not exist.
		NOTE:  This code uses i=0 as lowest level, which Holland calls
		level 1
*/

{
    int     kk, flag, pos;
    int     block, offset;
    unsigned mask = 1;
    unsigned tp;

    if (i == 0) {
	flag = 1;
	for (kk = 0; kk < SCHEMA; kk++) {
	    pos = kk + j * (SCHEMA + INTRON);
	    block = pos / UINTSIZE;
	    offset = pos % UINTSIZE;
	    tp = critter->chrom[block];
	    if (offset > 0)
		tp = tp >> offset;
	    if ((tp & mask) != 1)
		flag = 0;
	}
	if (flag == 1)
	    return (1);
	else
	    return (0);
    }
    else {
	if ((rr_func(critter, i - 1, 2 * j) == 1) &&
		(rr_func(critter, i - 1, 2 * j + 1) == 1))
	    return (1);
	else
	    return (0);
    }
}



void
app_set_options()
/* This routine is called once when a new run or extended run is started. */
/* It must set a value for all of its global variables, although they may */
/* be modified later (in other app_ functions) during the run. */
{
  
  avg_mut_width = 8;    /* Must set if using multimut.c.  Value is expected  */
			/* number of adjacent fields (bits, if binary) to    */
			/* change at once.  */

  stochastic = 0;       /* Must set to 0 or 1 to specify whether or not */
			/* fitness function always returns same value for a  */
			/* given chrom.-- If always same, set stochastic = 0 */
			/* to avoid extra evaluations; if changes with env't */
			/* or randomly, set = 1 */

  elitism = 1;          /* Must set to 0 or 1 to determine whether or not     */
			/* best indiv. is always preserved in next generation */
			/* (1 = yes).  */
  elitism &= !stochastic;   /* Note that elitism is reset to 0 automatically  */
			/* if user specifies stochastic flag = 1, since the   */
			/* changing fitness makes it impossible to insure that*/
			/* what was once best is still best. */

  user_supplied_initialization = 0;       /* Make it TRUE (1) if you will    */
					  /* supply code to initialize the   */
					  /* individuals in oldpop below, in */
					  /* routine app_user_init_pop(),    */
					  /* instead of using one of the     */
					  /* "standard" methods supplied with*/
					  /* GALOPPS.  */

  call_app_user_does_migration = 0;   /* Flag, if set, allows user full   */
					/* control of migration, from */
					/* app_user_does_migration(), below */

  using_inversion = 0;
			 /* Set flag to 1 (TRUE) if will use inversion to  */
			 /* alter placement of fields on chromosome, with  */
			 /* remapping of fields to "standard" positions at */
			 /* each evaluation, before objective function is  */
			 /* called.  It is set automatically if pinversion>0, */
			 /* but user may set manually here if has ever used */
			 /* inversion, so unmapping, translation, etc., will*/
			 /* continue to occur, even if pinversion now set  */
			 /* to 0.  */

  different_reps = 0 || using_inversion;
			 /* You set to 1 (TRUE) if will use different reps */
			 /* in different subpops.  If TRUE, and if the */
			 /* difference is NOT ONLY INVERSION, then YOU MUST*/
			 /* also supply code to transform ALL of the */
			 /* migrating individuals, in app_transform_migrant*/
			 /* in the skeleton below, and code below in */
			 /* app_transform_chrom_to_std to transform a */
			 /* chromosome into its STANDARD representation */
			 /* (used in the objective function objfunc().  */
			 /* Always 1 if inversion is being used.  */

  if (different_reps && !using_inversion) {
      /* User is using different representations in various subpopulations, */
      /* so must be supplying app_transform_chrom_to_std() and */
      /* app_transform_migrant().  User must also set the variables */
      /* max_chromsize_among_all_subpops and max_numfields_among_all_subpops */
      /* in the lines directly below, and delete the print and the call to */
      /* exit() which are inserted as a reminder to the user.  */ 

      fprintf(stderr, "\n*** You have selected different_reps and are not ");
      fprintf(stderr,"using inversion, \nso you MUST specify here in app_init");
      fprintf(stderr, " the maximum sizes of chromsize and numfields.***\n");
      exit(-9);
  }
  else {
      max_chromsize_among_all_subpops = -1; /* set to illegal value, as flag */
      max_numfields_among_all_subpops = -1; /* set to illegal value, as flag */
  }
}


void 
app_set_field_sizes()
/* Routine called ONLY IF user inputs alpha_size as < 2, meaning fields have */
/* different lengths.  In that case, user MUST provide code here to set the  */
/* array field_sizes[i], 0<=i<numfields, to the number of different codes    */
/* possible in the ith field (codes will vary from 0 to field_sizes[i] - 1). */
/* Data may be set in executable statements, or via initializers.   If the   */
/* problem uses different representations in different subpopulations, and   */
/* if that affects the field sizes, then this code must set the proper sizes */
/* for each subpopulation, using popno to tell which subpop is being         */
/* initialized.  Elsewhere, the program will also automatically use the      */
/* field_sizes to set array  field_ends, to the final bit number (in         */
/* [1, lchrom]) of each field.   */
{
  int i;

/*    field_sizes[0] = 24;     Replace with code to set them to the sizes  */
			    /* you actually want.  */
/*    field_sizes[9] = 48;     Replace with code to set them to the sizes  */
}



void
app_read_prob_params()
/* Application dependent data input, called ONCE at start of run (whether */
/* started from input parameters or EXTENDED from restart files) by main */
/* program BEFORE it reads the "standard" fields like numfields, */
/* numextrafields, lchrom, etc.  */
/* (Note: Onepop may do many RUNS, and this is called at start of each). */ 
/* Values read here are usually stored in static variables */
/* declared at the top of the app file, so they may be used in any of the */
/* user callback routines, but do NOT have to be read/written to the */
/* checkpoint files.  Thus, user should read these values ALSO when a run is */
/* restarted from saved checkpoint files.  If values MUST be preserved ACROSS */
/* runs (i.e., saved when checkpointing and reread when restarting) then */
/* that must be done in app_write_ckp_hdr() and app_read_ckp_hdr(), below, */
/* and the variables will be saved with EACH subpopulation.  Note: values */
/* like numfields, alpha_size, etc., are NOT yet known when this is called. */
{

}


void
app_data(REVISING)
BOOL    REVISING;
/* Application dependent data input, called by initdata() and revisedata(), */
/* just after user answers questions about crossover rate, scaling, */
/* convinterval, etc.  Ask input questions here which might be different for */
/* each subpopulation, and write (and read) the values to the checkpoint */
/* header file by putting code into app_write_ckp_hdr() and app_read_ckp_hdr()*/
/* below.  Values like numfields, alpha_size, etc., are already known when */
/* this routine is called. */
/* NOTE:  app_data is called whenever the user is initializing (from user */
/* input, flag REVISING == FALSE) or revising (flag REVISING == TRUE) input */
/* data (but only once if flag all_subpops_use_same_parameters is TRUE). */
/* HOWEVER, IF user is RESTARTING from saved checkpoint files, and does NOT */
/* ask to revise any parameters, this routing IS NOT CALLED.  Therefore, for */
/* user data (tables, etc.) which you want to read in on a RESTART, also, */
/* use app_init to read those data, as it is called whether restarting or */
/* starting from the beginning.  (Put it inside the "firstcall" logic.) */
{

    char    answer = 0;

    if (REVISING)
	return;
    /* Find out if want to print out special-format royal road chromosomes */
    if (interactive) {
	printf(" Do you want chromosomes printed each generation\n");
	printf(" in special royal_road format? (y/n) (%c)-->",
		print_rr_strings ? 'y' : 'n');
    }
    ffscanf("print_rr_strings", infp, "%c", &answer);
    if (answer == 0)
	return;
    else if (answer == 'y' || answer == 'Y')
	print_rr_strings = 1;
    else
	print_rr_strings = 0;
}


void
application()
/* this routine should contain any application-dependent computations */
/* that should be performed before each generation. called by main()    */
{
}


void
app_init()
/* Application dependent initialization routine called by initialize() for */
/* EACH subpopulation each cycle, whether started from input or restarted  */
/* from a restartfile, and whether all_subpops_use_same_parameters is TRUE */
/* or FALSE.  Called after individuals are malloc'd and after individuals  */
/* are read from .ind file (on a restart), but before random initialization*/
/* of all individuals (if initial start) or of individuals NOT read (if    */
/* restarting).                                                            */
{
  static int    firstcall = 1;
  int i;

  /* Any code you want to add to app_init goes below. */

    if (firstcall) {
	firstcall = 0;
	for (i = 0; i < 64; i++)        /* initialize array for watching rr */
					/* level changes */
	    savedrr_level[i] = 0;       /* i represents subpopulation number */

      /* Calculate max value of royal_road_level, for fitness function,etc. */
      max_royal_road_level = 0;
      for (i = 0; i < 5; i++) {
	if (power(i) == lchrom / (SCHEMA + INTRON))
	    max_royal_road_level = i;
	}
      if (max_royal_road_level < 0) {
	fprintf(outfp, "***BAD Chromosome Length... restart***/n");
      }
      first_time_stats = 1;
    }
}



void
app_user_init_pop(starting_guy_index)
/* If none of the standard methods supplied with GALOPPS will work for     */
/* initializing the individuals in oldpop at start of run, or for creating */
/* new "random" guys when popsize is expanded, or when convergence triggers*/
/* partial re-initialization, etc., then you can create the individuals    */
/* here, filling oldpop[starting_guy_index] through oldpop[popsize-1].     */
/* If you use this, set flag user_supplied_initialization in app_init().   */

int  starting_guy_index;
{
}


void
app_after_random_init()
/* Application-dependent changes or redoing of initialization, called after */
/* the random initialization is performed and (on restart) individuals have */
/* been read in from restart file. */
{
}


void
app_initreport()
/* Application-dependent initial report called by initialize() */
{

    skip(outfp, 1);
    fprintf(outfp, "Best Indiv. of Init. Pop., Roy Rd Lev. = %d, bonus = %5.2f,"
	    ,((struct rr_type *) (oldpop[bestnow].utility))->royal_road_level, ((struct rr_type *) (oldpop[bestnow].utility))->bonus);
    fprintf(outfp, " part = %5.2f\n", ((struct rr_type *) (oldpop[bestnow].utility))->part);
    fprintf(outfp, "Best Indiv. of new gen:\n");
    writechrom_rr(oldpop[bestnow].chrom);
    fprintf(outfp, "\n");
    skip(outfp, 1);
}



void
app_report()
/* Application-dependent report, called by report() */
{

	skip(outfp, 1);
	fprintf(outfp, "Best Indiv. of new gen., Royal Road Level = %d, bonus = %5.2f,"
		,((struct rr_type *) (newpop[bestnow].utility))->royal_road_level, ((struct rr_type *) (newpop[bestnow].utility))->bonus);
	fprintf(outfp, " part = %5.2f\n", ((struct rr_type *) (newpop[bestnow].utility))->part);
	fprintf(outfp, "Best Indiv. of new gen:\n");
	writechrom_rr(newpop[bestnow].chrom);
	fprintf(outfp, "\n");
	skip(outfp, 1);

	fprintf(outfp, "Best to date, Royal Road Level =%d, bonus = %5.2f,",
		((struct rr_type *) (bestfit.utility))->royal_road_level, ((struct rr_type *) (bestfit.utility))->bonus);
	fprintf(outfp, " part = %5.2f\n", ((struct rr_type *) (bestfit.utility))->part);

	fprintf(outfp, "Best individual to date:\n");
	writechrom_rr(bestfit.chrom);
	fprintf(outfp, "\n");
	skip(outfp, 1);
}


void
app_decide_if_converged(flag, num_to_replace_if_converged)
/* This routine is called at the end of each generation of each subpopulation */
/* to give the user the option of checking for convergence of the             */
/* subpopulation, or other condition.  When (if) that happens, if user wants  */
/* simply to reinitialize all or part of the subpopulation to random values   */
/* and keep running, user can just set *flag to 1 and set                     */
/* num_to_replace_if_converged to the desired number to reinitialize.  If     */
/* user wants to take some other action, should set *flag = 2 and code the    */
/* desired actions in routine app_when_converged() below.                     */
int * flag;
int * num_to_replace_if_converged;
{
/*  if(percent_converged > 0.xx) {
	*flag = 2;
	return;
    }
    if(avg_bias < 0.xx) {
	*flag = 2;
	return;
    }
    if(percent_converged > .xx) {
	*flag = 1;
	*num_to_replace_if_converged = popsize * .5;
	return;
    }
*/
}


void
app_when_converged()
/* This routine is called if user's routine (above) app_decide_if_converged  */
/* ever sets its flag to 2.  The intent is that user can here stop a run (by */
/* setting maxgen to the current value of gen, which will result in a        */
/* checkpoint file being written for possible restart), or change any GA     */
/* parameters as desired, for the current subpopulation, etc.  */
{
/*  if(avg_bias < .xx) {
	scalemult *= 1.xx;
	fprintf(outfp, "\n *** Scalemult raised xx%% for this pop, to %f\n", scalemult);
	return;
    }
	else {
	scalemult = (scalemult + 1.0) / 2.0;
	fprintf(outfp, "\n *** Scalemult lowered for this pop, to %f\n", scalemult);
	return;
    }
*/
}


void
app_quiet_report()
 /* Routine to be called when other output is suppressed, to check for
  * trigger for special user-defined output, or to print desired output even
  * when running in quiet mode.  Helpful when doing many long runs.
  * User can do checking for last generation of a Manypops run, for example,
  * in case want to reset quiet to 0 for a final summary printout, etc.   */

{
    int     rrlev;

    /* Below, supply any logic and printing to be used in quiet mode
     * operation, when most other normal output is suppressed. */

    /* for example, print warning ONE TIME if quiet flag is on, so user knows
     * why doesn't get any output.                                         */
    if(quiet < 3 && popno == 0 && cycle == 0 && gen == startgen)
	 fprintf(outfp, "\n Flag quiet > 0, normal output suppressed.\n");

    /* Logic below starts printing end of next-to-last generation of last
     * cycle, for final summary.  (Works for GALOPPS/Onepop (ncycles == 0) */
     /* and GALOPPS/Manypops.) */
    /* if((cycle == ncycles - 1 || ncycles == 0) && gen == maxgen - 2)
	  quiet = 0; */

    /* This logic causes print of information whenever the royal road     */
    /* level advances */
    rrlev = ((struct rr_type *) (bestfit.utility))->royal_road_level;

    if (rrlev > savedrr_level[popno] && rrlev > 0) {
	savedrr_level[popno] = rrlev;
	fprintf(outfp, "\n * Subpop %d, New rr_level = %d, neval = %ld, raw fitness = %f\n",
		popno, rrlev, bestfit.neval, bestfit.init_fitness);
	writechrom_rr(bestfit.chrom);
    }
}


void
app_new_global_best_report()
/* Application-dependent report, called by globalstats() in statisti.c when  */
/* GALOPPS/Manypops is run (not used by GALOPPS/Onepop runs).  It is called  */
/* after the LAST GENERATION of the cycle of EACH subpopulation.  That means */
/* that some things MAY have changed (inversion pattern, etc.) in some       */
/* generations BETWEEN when this best individual was found and now, when it  */
/* is to be printed, so user must be aware of that in writing outputs.  Also */
/* called in checkhdr.c to print the same values (local_bestfit.xxx) when   */
/* that local best turns out also to be the best among all pops, all        */
/* processes, so is EQUIVALENT at that time to printing all_pops_bestfit.xxx).*/
/* It should print out any app-dependent fields user will want to see.      */
/* (This call is not made at all from Onepop -- single-population runs).    */
{
  /* User can print from utility fields, call output routines, etc.        */
  /* to print whatever information is wanted when a new best individual of */
  /* all populations on all processes (processors) is found.  Since the    */
  /* performance measures, etc., have NOT yet been updated in the .stt     */
  /* file, user should now output local_bestfit.xxxxx, etc., not yet */
  /* all_pops_bestfit.xxxx, which will be updated AFTER this print is done.*/
  /* Use of this callback is OPTIONAL. */

    int i, rrlev;
    float bonus, part;

    rrlev = ((struct rr_type *) (local_bestfit.utility))->royal_road_level;
    bonus = ((struct rr_type *) (local_bestfit.utility))->bonus;
    part  = ((struct rr_type *) (local_bestfit.utility))->part;

    fprintf(outfp
		, "\n rr_level = %d, bonus = %f, part = %f \n"
		, rrlev, bonus, part);

    writechrom_rr(local_bestfit.chrom);
    /* report the number of schema 0 this ind has achieved */
    report_group_num(&local_bestfit);
    if (rrlev == 3) {
	fprintf(outfp, "\n Total evaluations to date, all subpops, = %ld\n",
		local_bestfit.neval);
	fprintf(outfp, "\n Run being terminated after success.");
	fflush(outfp);
	exit(1);
    }

}


report_group_num(critter)
struct individual * critter;
{
    int i;
    fprintf(outfp, "Groups present: ");
    for (i=0;i<16;i++)
	if (rr_func(critter,0,i)==1)
	    fprintf(outfp,"%d ",i+1);
    fprintf(outfp,"\n");
}


void
app_print_strings()
/* Optional application-dependent string printer, called by report() and */
/* initreport just after it has finished printing the chromosomes in */
/* "standard" binary format.  */
{

    if (print_rr_strings == 1) {
	repchar(outfp, " ", ((LINELENGTH - 17) / 2));
	fprintf(outfp, "Population Report\n");
	fprintf(outfp, "Generation %d\n", gen);

	fprintf(outfp, "num\n");
	fprintf(outfp, "old string\n");
	fprintf(outfp, "raw fitness | scaled fitness\n");

	if(gen != 0) {    /* FIX THIS TO IF INITIALIZING */
	    fprintf(outfp, "Generation %d\n", (gen + 1));
	    fprintf(outfp, "new string ");
	    skip(outfp, 1);
	    fprintf(outfp, "raw fitness | scaled fitness     ");
	    fprintf(outfp, "new string's parents, xsite(s) \n");
	    repchar(outfp, "-", LINELENGTH);
	    skip(outfp, 1);
	}
	writepop_rr();
	skip(outfp, 1);
    }
}


void
writepop_rr()
{
    struct individual *pind;
    int     j;

    for (j = 0; j < popsize; j++) {
	fprintf(outfp, "%3d) \n", j + 1);

	/* Old string */
	pind = &(oldpop[j]);
	writechrom_rr(pind->chrom);
	fprintf(outfp, "\n  %8f | %8f\n", pind->init_fitness, pind->fitness);

	if(gen != 0) {    /* FIX THIS TO IF INITIALIZING */
	    /* New string */
	    pind = &(newpop[j]);

	    writechrom_rr(pind->chrom);
	    fprintf(outfp, "\n  %8f | %8f     ", pind->init_fitness,
		    pind->fitness);
	    fprintf(outfp, "(%2d,%2d)   %2d  %2d  \n",
	      pind->parent[0], pind->parent[1], pind->xsite[0], pind->xsite[1]);
	}
    }
    return;
}

void
writechrom_rr(chrom)
/* Write a chromosome as a string of ones and zeroes            */
/* note that the most significant bit of the chromosome is the  */
/* RIGHTMOST bit, not the leftmost bit, as would be expected... */
unsigned *chrom;
{
    int     j, k, stop, pos;
    unsigned mask = 1, tmp;
    int     b1, b2;
    void    ruler();

    for (k = 0; k < chromsize; k++) {
	tmp = chrom[k];
	if (k == (chromsize - 1))
	    stop = lchrom - (k * UINTSIZE);
	else
	    stop = UINTSIZE;

	for (j = 0; j < stop; j++) {
	    if (tmp & mask)
		fprintf(outfp, "1");
	    else
		fprintf(outfp, "0");
	    tmp = tmp >> 1;
	    pos = k * UINTSIZE + j;

	    b1 = SCHEMA + INTRON;
	    b2 = 4 * b1;

	    if ((pos + 1) % b1 == 0)
		fprintf(outfp, " ");
	    if ((pos + 1) % b2 == 0)
		fprintf(outfp, "\n");

	}
    }
    ruler(outfp);
    return;
}


void
app_conv_rept()
/* Application-dependent report, called by reptconv() (in report.c) */
{

}


BOOL
app_user_does_migration(IndArray, filename, number_to_migrate, want_best
				, migration_crowding_amount, ChromSize
				, migration_incest_reduction, BestInd
				, migrant_holder)
/* In this routine, if the user has set the global variable (in app_init)   */
/* call_app_user_does_migration, and user is NOT using inversion, then the  */
/* user HERE is allowed to perform migration among subpopulations in any    */
/* manner the user chooses.  ChromSize is the number of unsigneds in the    */
/* chromosome of the DONOR population, and global chromsize is the number of*/
/* unsigneds in the chromosome of the RECEIVING population (in IndArray).   */
/* IndArray is the receiving population (the global variable is oldpop),    */
/* and transfer occurs at the start of a cycle); the index of the best      */
/* individual in oldpop (in internal form, without one added) is in BestInd.*/
/* For guidance in performing the required operations, see routine          */
/* GetMigrants() in file checkrd.c, where the system-provided migration     */
/* capabilities (migration_incest_reduction, migration crowding, inversion, */
/* etc., are all done.                                                      */
struct individual *IndArray;
char   *filename;
int    number_to_migrate;
BOOL   want_best;
int    migration_crowding_amount, ChromSize, migration_incest_reduction;
int    BestInd;
struct individual *migrant_holder;
{
    return TRUE;
}

BOOL
app_transform_chrom_to_std(mapped_saved_chromosome, chrom_for_objfunc)
/* If the user is using different representations for different subpops,    */
/* but is NOT using the system-provided inversion code, here is where the   */
/* user must supply the code to transform a migrant individual's chromosome */
/* (in mapped_saved_chromosome) from its "mapped" representation in the */
/* subpopulation to the STANDARD representation, which is the one the user */
/* uses in defining the objective (or fitness) function, objfunc.  The */
/* "standard rep" chromosome must be returned in chrom_for_objfunc, which is */
/* a local name for critter->chrom, the chromosome passed to objfunc().  */
/* The user can use any of the global variables, or local variables defined */
/* at the top of this file, to determine which supopulation this is (popno), */
/* and what characteristics THIS representation may have (local variables, */
/* utility fields, etc.) */

unsigned *mapped_saved_chromosome;
unsigned *chrom_for_objfunc;
{
    int ii;

    /* The commented-out code below does the "identity" transformation on the */
    /* chromosome -- i.e., copies one to the other. If the user is doing some */
    /* transformation OTHER than the system-provided inversion, user */
   /* should change the first loop below to perform the actual */
    /* transformation process.  Finally, user must make the function return */
    /* TRUE if successful. */

    /* NOTE:  Next loop must be modified if subpops have different chromosome */
    /* representations -- user must do a transformation, not a copy.  */
/*  for (ii=0;ii<chromsize;ii++) {
	chrom_for_objfunc[ii] = mapped_saved_chromosome[ii];
    }
    return TRUE;
*/
    return FALSE;
}



BOOL
app_transform_migrant(migrant_holder, DonorChromSize, newguy, filename)
/* If the user is using different representations for different subpops,    */
/* but is NOT using the system-provided inversion code, here is where the   */
/* user must supply the code to transform a migrant individual (in a struct */
/* individual pointed to by migrant_holder) from its donor pop rep to its   */
/* popno (a global, the current "receiving" population) representation,     */
/* returning the transformed guy in newguy, which has a chromosome length of*/
/* RecipientChromSize unsigneds.   If user is using inversion, not user's   */
/* "own" transformation, then this routine is not called, so may be left in */
/* its "commented out" form, which is the the IDENTITY transformation (i.e.,*/
/* copy the individual unchanged from migrant_holder to newguy.             */

/* A pointer to the individual to migrate, as read from donorpop, and its   */
/* DonorChromSize, are provided, as is the filename of the recipient pop.   */
/* After transforming the individual, the user should place it into new_guy */
/* and return TRUE if successful.            */
struct individual *migrant_holder, *newguy;
int    DonorChromSize;
char   *filename;
{
    int ii;
/* The commented-out code below does the "identity" transformation.  If user  */
/* is doing some transformation OTHER than the system-provided inversion, user*/
/* should change the first loop below to perform the actual transformation    */
/* process.  Finally, make the function return TRUE if successful. */

/* NOTE:  limit chromsize in next loop below must be revised if subpops have */
/* different chromosome sizes... user must do a transformation, not a copy.  */
/*  for (ii=0;ii<chromsize;ii++) {
	newguy->chrom[ii] = migrant_holder->chrom[ii];
    }
    newguy->parent[0] = migrant_holder->parent[0];
    newguy->parent[1] = migrant_holder->parent[1];
    newguy->xsite[0] = migrant_holder->xsite[0];
    newguy->xsite[1] = migrant_holder->xsite[1];
    newguy->fitness = migrant_holder->fitness;
    newguy->init_fitness = migrant_holder->init_fitness;
    newguy->neval = migrant_holder->neval;
    sys_util_copy_utility(newguy->utility, migrant_holder->utility);

    return TRUE;
*/
    return FALSE;
}


void
app_generate()
/* Opportunity to do any desired (application-dependent) operations at the */
/* end of each generation of each subpopulation. */
{
  /* Nothing to do for this application. */
}



void
app_stats(pop)
/* Application-dependent statistics calculations called by statistics() at  */
/* the end of each generation.  */

struct individual *pop;
{
    int i, j, ngps;

    if (first_time_stats) {
	ngps = lchrom/15;
	for (j=0;j<ngps;j++)
	    group_rec[j] = 0;
	for (i=0;i<popsize;i++)
	    for (j=0;j<ngps;j++)
		if (group_rec[j]==0)
		    group_rec[j] = rr_func(&(pop[i]),0,j);
	fprintf(outfp, "\nGroups present in subpop %d: ", popno);
	for (j=0;j<ngps;j++)
	    fprintf(outfp, "%d ",group_rec[j]);
	fprintf(outfp, "\n");
	first_time_stats = 1;
    }
}


void
app_before_inversion()
/* Allow user to perform any desired application-specific actions before */
/* calling the inversion routine.  Only called when an inversion WILL be */
/* performed on the current subpopulation at the current generation. */
{

}



BOOL
app_write_ckp_hdr(fp)
FILE   *fp;

/* Application-dependent callback routine user may use for writing      */
/* any needed application-specific fields into a checkpoint file.       */
{
    if (fwrite(&max_royal_road_level, sizeof(int), 1, fp) != 1)
	return FALSE;
    if (fwrite(&print_rr_strings, sizeof(int), 1, fp) != 1)
	return FALSE;
    return TRUE;
}


BOOL
app_read_ckp_hdr(fp)
FILE   *fp;
{
    if (fread(&max_royal_road_level, sizeof(int), 1, fp) != 1)
	return FALSE;
    if (fread(&print_rr_strings, sizeof(int), 1, fp) != 1)
	return FALSE;
    return TRUE;
}



int
GetUtilitySize()
/* User must here return 0 (utility fields not used) or the number of bytes, */
/* sizeof(struct(WhateverUtilityFieldsYouUse)) of utility fields used. */
{
    int nbytes;

    nbytes = sizeof(struct rr_type);
    return nbytes;
}


int
GetMaxUtilitySize()
/* User must here return 0 (utility fields not used) or GetUtilitySize (if */
/* different_reps are NOT used), or the size (bytes) of the LARGEST utility */
/* field used by ANY subpopulation, if different_reps includes use of */
/* utility fields of different sizes. */
{
  int nbytes;  
  
  nbytes = GetUtilitySize();
  if (nbytes) {
    if (different_reps) 
      /* Here, return size of largest utility field used by any rep. */ 
      return nbytes;    /* (for royal road problem, always same size) */
    else
      return nbytes;   
  }  
  return nbytes;
}


void
ruler(outfp)
FILE   *outfp;
{
    int     i, j;

    for (i = 0; i < 4; i++) {
	for (j = 0; j < SCHEMA; j++) {
	    fprintf(outfp, "-");
	}
	for (j = 0; j < INTRON; j++) {
	    fprintf(outfp, "+");
	}
	fprintf(outfp, " ");
    }
    fprintf(outfp, "\n");
}
