/*----------------------------------------------------------------------------*/
/* bitmutat.c -- Single-bit and field-wise mutation operators for ordinary    */
/*               (non-permutation)                                            */
/*               problems.   Mutates each locus (field) with probability      */
/*               pmutation, the prob. of mut. per field (i.e., per BIT, if a  */
/*               binary problem; otherwise, per field, producing a legal code */
/*               within the specified alpha_size or field_lengths[i]).        */
/*               Modified 3/95 by EDG to speed up, avoid loops.               */
/*               Modified 4/96 by EDG to handle fields with different lengths */
/*               (Also uses new getfield, putfield to do it. */
/*----------------------------------------------------------------------------*/

#include "external.h"
#include <math.h>

char which_mutation[] = "bitmutat.c";

void
mutation(child)
unsigned *child;

/* Mutate each allele (bit) with probability pmutation, count # of mutations */
/* Actually implemented by using derived probability of 1 or more mutations  */
/* on this chromosome, flipping to see if get one, if so, doing it, flipping */
/* again, continuing until DON'T get one, for reasonable mutation rates (see */
/* below). */
{
    int mut_locus, wordnumber, bitinword;
    unsigned int mask;
    /* Make next two static and test so only compute p_anymuts once for any */
    /* given pmutation. */
    int i;
    static float p_anymuts;
    static double local_pmutation;

    if(local_pmutation != pmutation) {
        local_pmutation = pmutation;
	if (local_pmutation >= 1.0) {
	    fprintf(stderr, "\n*** Bad mutation rate specified, exiting.\n");
	    exit(-1);
	}
	p_anymuts = 1.0
		    - (float)pow((double)(1.0 - pmutation), (double)lchrom);
    }
    if (p_anymuts < 0.8) {
      /* Will do [none - few] mutations/chromosome, so it is reasonable to */
      /* approximate the process by sampling with replacement, just doing */
      /* mutations at random locations until flip (probability of further */
      /* mutations) comes up "no".  This is quite fast. */
 
      while(flip(p_anymuts)) {
        mut_locus = rnd(0, (lchrom - 1));
        wordnumber = mut_locus/UINTSIZE;
	bitinword = mut_locus%UINTSIZE;
	mask = 1 << bitinword;
	child[wordnumber] ^= mask;
	nmutation++;
      }
    }
    else {
      /* In this case, sampling with replacement not a very good approx., */
      /* so just go along chromosome and,                                 */
      /* at each locus, with probability pmutation, do a mutation there.  */
      /* NOTE:  this case generally represents a MISUSE of the mutation */
      /* operator, in the GA context, as mutation will likely overwhelm */
      /* selection and reproduction, causing "scrambling" of the information, */
      /* and lack of ability to learn and converge.  But this will prevent */
      /* "infinite" loops in the code from oversized mutation probabilities. */
      for (i=0;i<numfields;i++) {
        if (flip(pmutation)) {
          wordnumber = i/UINTSIZE;
	  bitinword = i%UINTSIZE;
	  mask = 1 << bitinword;
	  child[wordnumber] ^= mask;
	  nmutation++;
        }
      }

    }
}

void
field_mutation(child)
unsigned *child;

/* Mutate chosen fields to uniform randomly chosen members of those fields'  */
/* legal alphabets, with probability pmutation; count # of mutations         */
/* Actually implemented by using derived probability of 1 or more mutations  */
/* on this chromosome, flipping to see if get one, if so, doing it, flipping */
/* again, continuing until DON'T get one. */
/* NOTE:  probability of mutation is same for each field, regardless of its  */
/* length (not realistic in nature, but usual in GA).                        */
{
    int mut_field, allele, new_allele, i;
    /* Make next two static and test so only compute p_anymuts once for any */
    /* given pmutation. */
    static float p_anymuts;
    static double local_pmutation;

    if(local_pmutation != pmutation) {
	local_pmutation = pmutation;
	if (local_pmutation >= 1.0) {
	    fprintf(stderr, "\n*** Bad mutation rate specified, exiting.\n");
	    exit(-1);
	}
	p_anymuts = 1.0
		    - (float)pow((double)(1.0 - pmutation), (double)numfields);
    }
    if (p_anymuts < 0.8) {
      /* Will do [none - few] mutations/chromosome, so it is reasonable to */
      /* approximate the process by sampling with replacement, just doing */
      /* mutations at random locations until flip (probability of further */
      /* mutations) comes up "no".  This is quite fast. */

      while(flip(p_anymuts)) {
	mut_field = rnd(0, (numfields - 1));
	allele = getfield(mut_field, numfields, child);
	while ((new_allele = rnd(0, field_sizes[mut_field] - 1)) == allele) {
	}
	putfield(mut_field, numfields, new_allele, child);

	nmutation++;
      }
    }
    else {
      for (i=0;i<numfields;i++) {
        if (flip(pmutation)) {
	  allele = getfield(i, numfields, child);
	  while ((new_allele = rnd(0, field_sizes[i] - 1)) == allele) {
  	  }
	  putfield(i, numfields, new_allele, child);

	  nmutation++;
        }
      } 
    }
}
