/*---------------------------------------------------------------------------*/
/* checkhdr.c -- routines for reading and writing checkpoint and restart     */
/*                  files with suffix .ckp.  Also reads/writes global stats  */
/*                  files (.stt), if multiple processes are involved in a run*/
/*                  or set of runs.                                          */
/* NOTE:  If you make any changes to the file formats below, you MUST also   */
/* record the effect of those changes in the defines in file master.c, or    */
/* the inversion operator WILL NOT WORK and ALL GALOPPS operations will be   */
/* severely compromised.                                                     */
/*---------------------------------------------------------------------------*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "external.h"

extern int local_bestfit_popno;
extern int local_bestfit_chromsize;
extern int local_bestfit_numfields;

static char checkpt_stt_filename[13];
static char restart_stt_filename[13];
static char buffer[8];

static int All_Pops_Bestfit_Chromsize, All_Pops_Bestfit_Popno;
static int All_Pops_Bestfit_Numfields;

#define RETRIES 5

BOOL
WriteCheckPointHeader(fname, CallBackFun)
     char   *fname;

     BOOL(*CallBackFun) ( /* FILE * */ );
{
  FILE   *fp;
  int     i;
  long states[2];


  if (!(fp = start_write(fname, states, RETRIES))) {
    fprintf(stderr, "\n*** ERROR: Unable to open checkpoint header file %s\n",
	    fname);
    fprintf(stderr, " for writing.  Will terminate run. ***\n");
    return FALSE;
  }

  if (fwrite(&popsize, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }

  /* For this full checkpoint, wrote all indiv to file, so popsize appears */
  /* twice in file... but for other applications, number below may be  */
  /* less than popsize... it will be READ as the number of individuals */
  /* which were written to the file   */
  if (fwrite(&popsize, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }

  if (fwrite(&chromsize, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&numfields, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }

/* CAUTION:  Putting any more writes in the top part of this checkpoint */
/* header file (before field_sizes is written) may require changing */
/* the #defines at top of file master.c. */

  if (fwrite(&fieldlength, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  /* New array added at Release 2.50, for inversion map */
  for (i=0;i<numfields;i++) {
    if (fwrite(&(this_pop_map_to_std[i]), sizeof(int), 1, fp) != 1) {
      fclose(fp);
      return FALSE;
    }
  }
  /* Next two added June, 1996, EDG */
  if (fwrite(&max_chromsize_among_all_subpops, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&max_numfields_among_all_subpops, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  /* field_ends and field_sizes added 4/96 by EDG */
  for (i=0;i<numfields;i++) {
    if (fwrite(&(field_ends[i]), sizeof(int), 1, fp) != 1) {
      fclose(fp);
      return FALSE;
    }
  }
  /* (also added 4/96, EDG) */
  for (i=0;i<numfields;i++) {
    if (fwrite(&(field_sizes[i]), sizeof(int), 1, fp) != 1) {
      fclose(fp);
      return FALSE;
    }
  }
  if (!(fwrite(bestfit.chrom, sizeof(unsigned), 
	       max_chromsize_among_all_subpops, fp))) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&(bestfit.fitness), sizeof(double), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&(bestfit.init_fitness), sizeof(double), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&(bestfit.neval), sizeof(long), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&(bestfit.generation), sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  for (i=0;i<max_numfields_among_all_subpops;i++) {
    if (fwrite(&(bestfit.map_to_std[i]), sizeof(int), 1, fp) != 1) {
      fclose(fp);
      return FALSE;
    }
  }
  
  for (i=0;i<max_numfields_among_all_subpops;i++) {
    if (fwrite(&(bestfit.field_ends[i]), sizeof(int), 1, fp) != 1) {
      fclose(fp);
      return FALSE;
    }
  }

  if (sys_util_write_utility(bestfit.utility, fp) == FALSE) {
    fclose(fp);
    return FALSE;
  }

  for (i=0;i<max_numfields_among_all_subpops;i++) {
    if (fwrite(&(std_field_ends[i]), sizeof(int), 1, fp) != 1) {
      fclose(fp);
      return FALSE;
    }
  }

  if (fwrite(&one_pop_cum_sumfitness, sizeof(double), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&one_pop_sum_best_fitness, sizeof(double), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&one_pop_best_fitness_count, sizeof(long), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&one_pop_online_denominator, sizeof(long), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&minraw, sizeof(double), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&avgraw, sizeof(double), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&fit_max, sizeof(double), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&pcross, sizeof(float), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&pmutation, sizeof(float), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  /* Next 2 fields added 6/96, EDG. */
  if (fwrite(&p_multimut, sizeof(float), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&avg_mut_width, sizeof(float), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&pinversion, sizeof(float), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  /* Next 3 fields added to header in April, 96 release, EDG */
  if (fwrite(&autocontrol_selection, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&beta, sizeof(double), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&beta_s_tilde, sizeof(float), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&scalemult, sizeof(float), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&conv_measure, sizeof(float), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&conv_sigma_coeff, sizeof(float), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&sigma_trunc, sizeof(float), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&scaling_window, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&windowstart, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&windowend, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  for (i = 0; i < MAXSCALINGWINDOW; i++)
    if (fwrite(&(savedmins[i]), sizeof(double), 1, fp) != 1) {
      fclose(fp);
      return FALSE;
    }
  if (fwrite(&sigma, sizeof(float), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&neval, sizeof(long), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&lchrom, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&genspercycle, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&gen, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&maxgen, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&printstrings, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&nmutation, sizeof(long), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&ncross, sizeof(long), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&bestnow, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&convinterval, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&crowding_factor, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }

  if (fwrite(&incest_reduction, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }

  if (fwrite(&stochastic, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&elitism, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }


  for (i = 0; i < 55; i++)
    if (fwrite(&(oldrand[i]), sizeof(double), 1, fp) != 1) {
      fclose(fp);
      return FALSE;
    }
  if (fwrite(&jrand, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&rndx2, sizeof(double), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fwrite(&rndcalcflag, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  /* New global added at Release 2.30, alpha_size */
  if (fwrite(&alpha_size, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }

  /* New global added to saved variables at Release 2.35, permproblem */
  if (fwrite(&permproblem, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }

  /* New global added to saved variables at Release 2.37, tourneysize */
  /* (always written/read; used only if tournament selection is in use.) */
  if (fwrite(&tourneysize, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }

  /* Calls user-defined function to write any state information added by */
  /* the user.  */
  if ((*CallBackFun) (fp) == FALSE) {
    fclose(fp);
    return FALSE;
  }

  return end_write(fname, fp, states, RETRIES);
}


int
ReadCheckPointHeader(fname, CallBackFun)
     char   *fname;

     BOOL(*CallBackFun) ( /* FILE * */ );
{
  FILE   *fp;
  int     indcnt, i, buf_pos;
  unsigned long mask, index;
  unsigned nbytes;

  BOOL read_complete;

  if (!(fp = start_read(fname, 1, RETRIES, &mask, &index, &buf_pos))) {
    fprintf(stderr, "\n*** ERROR: Unable to open %s for read ***\n", fname);
    fprintf(stderr, "\nRun will be terminated.\n");
    return -1;                  /* Flag for failure */
  }

  if (fread(&popsize, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }

  /* indcnt, the number of individuals written to this file, need not */
  /* match popsize when it was run -- but will for full checkpoint files.  */
  if (fread(&indcnt, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }

  if (fread(&chromsize, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&numfields, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  /* numextrafields eliminated 4/96 by EDG */

  if (fread(&fieldlength, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  
  /* Before can read this in, must use the value for numfields just read */
  /* to malloc the proper amount of storage for this map. */
  nbytes = numfields*sizeof(int);
  if ((this_pop_map_to_std = (int *) malloc(nbytes)) == NULL)
    nomemory("this_pop_map_to_std");
  for (i=0;i<numfields;i++) {
    if (fread(&(this_pop_map_to_std[i]), sizeof(int), 1, fp) != 1) {
      fclose(fp);
      return -1;
    }
  }
  /* Next two added June, 1996, EDG */
  if (fread(&max_chromsize_among_all_subpops, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }
  if (fread(&max_numfields_among_all_subpops, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return FALSE;
  }

  /* field_sizes and field_ends added 4/96 by EDG */
  malloc_field_sizes_arrays();
  for (i=0;i<numfields;i++) {
    if (fread(&(field_ends[i]), sizeof(int), 1, fp) != 1) {
      fclose(fp);
      return -1;
    }
  }
  for (i=0;i<numfields;i++) {
    if (fread(&(field_sizes[i]), sizeof(int), 1, fp) != 1) {
      fclose(fp);
      return -1;
    }
  }

  /* If max sizes for bestever variables not yet set, must do here, now */
  /* that we know chromsize and numfields. (If using different_reps and */
  /* not using inversion, these will already be set in app_set_options.)  */
  /* (Do only FIRST time through -- never freed. */
  if (max_chromsize_among_all_subpops == -1)
    max_chromsize_among_all_subpops = chromsize;

   if (max_numfields_among_all_subpops == -1)
     max_numfields_among_all_subpops = numfields;

   /* Now, can get the variable-length fields in bestever */
   /* structs malloc'd  (NOP after first call). */
   bestever_variable_malloc();
   if (fread(bestfit.chrom
	     , (size_t) max_chromsize_among_all_subpops * sizeof(unsigned)
	     , 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&(bestfit.fitness), sizeof(double), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&(bestfit.init_fitness), sizeof(double), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&(bestfit.neval), sizeof(long), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&(bestfit.generation), sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  for (i=0;i< max_numfields_among_all_subpops;i++) {
    if (fread(&(bestfit.map_to_std[i]), sizeof(int), 1, fp) != 1) {
      fclose(fp);
      return FALSE;
    }
  }
  for (i=0;i< max_numfields_among_all_subpops;i++) {
    if (fread(&(bestfit.field_ends[i]), sizeof(int), 1, fp) != 1) {
      fclose(fp);
      return FALSE;
    }
  }
  if (sys_util_read_utility(bestfit.utility, fp) == FALSE) {
    fclose(fp);
    return -1;
  }

  /* added 5/96, EDG */
  for (i=0;i< max_numfields_among_all_subpops;i++) {
    if (fread(&(std_field_ends[i]), sizeof(int), 1, fp) != 1) {
      fclose(fp);
      return FALSE;
    }
  }

  if (fread(&one_pop_cum_sumfitness, sizeof(double), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&one_pop_sum_best_fitness, sizeof(double), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&one_pop_best_fitness_count, sizeof(long), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&one_pop_online_denominator, sizeof(long), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&minraw, sizeof(double), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&avgraw, sizeof(double), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&fit_max, sizeof(double), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }

  if (fread(&pcross, sizeof(float), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&pmutation, sizeof(float), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  /* Next 2 fields added 6/96, EDG. */
  if (fread(&p_multimut, sizeof(float), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&avg_mut_width, sizeof(float), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&pinversion, sizeof(float), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  /* Next 3 fields added to header in April, 96 release, EDG */
  if (fread(&autocontrol_selection, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&beta, sizeof(double), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&beta_s_tilde, sizeof(float), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&scalemult, sizeof(float), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }

  if (fread(&conv_measure, sizeof(float), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&conv_sigma_coeff, sizeof(float), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&sigma_trunc, sizeof(float), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&scaling_window, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&windowstart, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&windowend, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  for (i = 0; i < MAXSCALINGWINDOW; i++)
    if (fread(&savedmins[i], sizeof(double), 1, fp) != 1) {
      fclose(fp);
      return -1;
    }
  if (fread(&sigma, sizeof(float), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&neval, sizeof(long), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&lchrom, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&genspercycle, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&gen, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&maxgen, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&printstrings, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&nmutation, sizeof(long), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&ncross, sizeof(long), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&bestnow, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&convinterval, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&crowding_factor, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }

  if (fread(&incest_reduction, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }

  if (fread(&stochastic, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&elitism, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  for (i = 0; i < 55; i++)
    if (fread(&(oldrand[i]), sizeof(double), 1, fp) != 1) {
      fclose(fp);
      return -1;
    }
  if (fread(&jrand, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&rndx2, sizeof(double), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if (fread(&rndcalcflag, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }

  /* New global variable added at Release 2.30, alpha_size */
  if (fread(&alpha_size, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  /* New global added to saved variables at Release 2.35, permproblem */
  if (fread(&permproblem, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  /* New global added to saved variables at Release 2.37, tourneysize */
  /* (always written/read; used only if tournament selection is in use.) */
  if (fread(&tourneysize, sizeof(int), 1, fp) != 1) {
    fclose(fp);
    return -1;
  }
  if ((*CallBackFun) (fp) != TRUE) {
    fclose(fp);
    return -1;
  }

  if ((end_read(fname, fp, 1, RETRIES, mask, index, buf_pos, 
	       &read_complete)) && read_complete == TRUE)
    return indcnt;
  
  return -1;
}


BOOL
ReadCumStats()
     /* Called the FIRST time a processor is restarted from saved checkpoint */
     /* files or from keyboard or file initialization.  Subsequent calls are */
     /* instead made to UpdateCumStats().  Difference is that this one is    */
     /* called at a time when we're NOT ready to update (write to) the file  */
     /* after reading it. However, if it doesn't exist, and is the beginning */
     /* of a new run (gen == 0), and if the processor which "owns" subpop 0  */
     /* finds it missing, it will write a 0'd file to initialize the whole   */
     /* run. If it doesn't exist and it is NOT gen == 0 and the processor    */
     /* owns subpop 0, then the contents of the 'restartfileprefix'99.stt is */
     /* copied to 'checkptfileprefix'99.stt to initialize it.                */
     /* Note that all is bypassed if single_process is TRUE, as stt file is  */
     /* not needed if only one process will be used. */ 
{
  FILE *stt_fp;

  int MustCopyRestartSttToCheckptStt = 0;
  int i;
  static int firstcall = 1;

  if (single_process) return TRUE;

  /* If this process "owns" subpop 0 (i.e., startpopnum == 0), then it has */
  /* the duty to zero the cumulative statistics, which it will keep in file*/
  /* 'checkptfileprefix'99.stt, at the beginning of the run, or to copy it */
  /* from the previous run's .stt file, if not the beginning of  the run.  */
  /* If the new file already exists at the beginning of a run (gen == 0), */
  /* then the values in it will be thrown out,        and the file will be */
  /* initialized to zeros, informing the user that this is happening.      */
  if (firstcall == 1) {
    if (!strlen(checkptfileprefix) && !strlen(restartfileprefix))
      strcpy(buffer, "sgackp");
    else if (!strlen(checkptfileprefix))
      strcpy(buffer, restartfileprefix);
    else
      strcpy(buffer, checkptfileprefix);
    CombineFileName(checkpt_stt_filename, buffer, 99, ".stt");
  }
  if (startpopnum != 0 && gen == 0)
    return TRUE;
  if(startpopnum == 0 && firstcall == 1) {
    firstcall = 0;
    if (gen == 0) {
      if ((stt_fp = fopen(checkpt_stt_filename, "rb")) != NULL) {
	fclose(stt_fp);
	remove(checkpt_stt_filename);
	fprintf(stderr,"\nRun started with an existing %s file "
		,checkpt_stt_filename);
	fprintf(stderr,"which was probably garbage.\nThe cumulative ");
	fprintf(stderr, "statistics were reinitialized to zero.\n");
	fprintf(outfp,"\nRun started with an existing %s file "
		, checkpt_stt_filename);
	fprintf(outfp,"which was probably garbage.\nThe cumulative ");
	fprintf(outfp, "statistics were reinitialized to zero.\n");
      }
      if(!InitializeCumStats()) {
	fprintf(stderr, "\n***ERROR Couldn't initialize %s."
		, checkpt_stt_filename);
	exit(-1);
      }
      return TRUE;
    }
    else {
      /* Not gen 0, so must have a previous 'restartfileprefix'99.stt */
      /* file which we must copy to 'checkptfileprefix'99.stt.  */
      if (Lock(checkpt_stt_filename, 10) == FALSE) {
	fprintf(stderr, "\nCan't lock %s.  Will set global stats"
		, checkpt_stt_filename);
	fprintf(stderr,"\nto 0.  ***NOTE:  Means the global stats that follow");
	fprintf(stderr, "are NOT CORRECT!!!\n  They omit cum. statistics ");
	fprintf(stderr, "BEFORE this restart was initiated.  \n You must ");
	fprintf(stderr, "correct them yourself. ***");
	fprintf(outfp, "\nCan't lock %s.  Will set global stats"
		, checkpt_stt_filename);
	fprintf(outfp, "\nto 0.  ***NOTE:  Means global stats that follow ");
	fprintf(outfp, "are NOT CORRECT!!!\n  They omit cum. statistics ");
	fprintf(outfp, "BEFORE this restart was initiated.  \n You must ");
	fprintf(outfp, "correct them yourself. ***");
	if(!InitializeCumStats()) {
	  fprintf(stderr, "\n***ERROR Couldn't initialize %s."
		  , checkpt_stt_filename);
	  exit(-1);
	}
      }
      if (strcmp(restartfileprefix, checkptfileprefix) != 0)
	MustCopyRestartSttToCheckptStt = 1;
      /* Must copy the 'restartfileprefix'99.stt file, later, after */
      /* read it, to make new 'checkptprefix'99.stt */
    }
  }
  /* checkpt_stt_file locked, gen == 0 if here with startpopnum == 0 */
  if (startpopnum != 0 && gen != 0 && strlen(restartfileprefix)) {
    CombineFileName(restart_stt_filename, restartfileprefix, 99, ".stt");
    if (Lock(restart_stt_filename, 20) == FALSE) {
      fprintf(stderr, "\n*** ERROR. Couldn't lock %s in 20 tries"
	      , restart_stt_filename);
      fprintf(stderr, "\n*** Ignore global statistics for first cycle");
      fprintf(stderr, " of this process.\n");
      fprintf(outfp, "\n*** ERROR. Couldn't lock %s in 20 tries"
	      , restart_stt_filename);
      fprintf(outfp, "\n*** Ignore global statistics for first cycle");
      fprintf(outfp, " of this process.\n");
      return TRUE;
    }
    /* It locked okay, if got here. */
  }
  if (startpopnum == 0 && gen != 0) {
    CombineFileName(restart_stt_filename, restartfileprefix, 99, ".stt");
    if (Lock(restart_stt_filename, 20) == FALSE) {
      fprintf(stderr, "\n*** ERROR. Couldn't lock %s in 20 tries"
	      , restart_stt_filename);
      fprintf(stderr, "\n*** Ignore global statistics:  data from ");
      fprintf(stderr, " restart was lost.\n");
      fprintf(outfp, "\n*** ERROR. Couldn't lock %s in 20 tries"
	      , restart_stt_filename);
      fprintf(outfp, "\n*** Ignore global statistics:  data from ");
      fprintf(outfp, " restart was lost.\n");
      return TRUE;
    }
    /* It locked okay, if got here. */
  }
  /* Okay, the file locked okay, so try to open and get its contents. */
  if ((stt_fp = fopen(restart_stt_filename, "rb")) == NULL) {
    fprintf(stderr, "\nLocked file %s didn't exist.", restart_stt_filename);
    fprintf(stderr, "\nDelete its lock file z?????99.stt (if any) and "); 
    fprintf(stderr, "redo the run\nwhich should create it.\n");
    exit(-1);
  }

  /* Find out length of chromosome, etc., for population whose statistics */
  /* are recorded as all_pops numbers here, so can malloc space for reading */
  /* them in. */
  if (fread(&(All_Pops_Bestfit_Popno), sizeof(int), 1, stt_fp) != 1) {
    fclose(stt_fp);
    return FALSE;
  }
  if (fread(&(All_Pops_Bestfit_Chromsize), sizeof(int), 1, stt_fp) != 1) {
    fclose(stt_fp);
    return FALSE;
  }
  if (fread(&(All_Pops_Bestfit_Numfields), sizeof(int), 1, stt_fp) != 1) {
    fclose(stt_fp);
    return FALSE;
  }

  /* Get the variable-length fields in bestever structs malloc'd */
  bestever_variable_malloc();

  if (fread(&all_pops_sumfitness, sizeof(double), 1, stt_fp) != 1) {
    fclose(stt_fp);
    return FALSE;
  }
  if (fread(&all_pops_sum_best_fitness, sizeof(double), 1, stt_fp) != 1) {
    fclose(stt_fp);
    return FALSE;
  }
  if (fread(&all_pops_online_denominator, sizeof(long), 1, stt_fp) != 1) {
    fclose(stt_fp);
    return FALSE;
  }
  if (fread(&all_pops_best_fitness_count, sizeof(long), 1, stt_fp) != 1) {
    fclose(stt_fp);
    return FALSE;
  }
  if (fread(&all_pops_neval, sizeof(long), 1, stt_fp) != 1) {
    fclose(stt_fp);
    return FALSE;
  }
  if (fread(all_pops_bestfit.chrom
	    , All_Pops_Bestfit_Chromsize * sizeof(unsigned)
	    , 1, stt_fp) != 1) {
    fclose(stt_fp);
    return FALSE;
  }
  if (fread(&(all_pops_bestfit.fitness), sizeof(double), 1, stt_fp) != 1) {
    fclose(stt_fp);
    return FALSE;
  }
  if (fread(&(all_pops_bestfit.init_fitness), sizeof(double),1,stt_fp) != 1) {
    fclose(stt_fp);
    return FALSE;
  }
  if (fread(&(all_pops_bestfit.neval), sizeof(long), 1, stt_fp) != 1) {
    fclose(stt_fp);
    return FALSE;
  }
  if (fread(&(all_pops_bestfit.generation), sizeof(int), 1, stt_fp) != 1) {
    fclose(stt_fp);
    return FALSE;
  }
  for (i=0;i<All_Pops_Bestfit_Numfields;i++) {
    if(fread(&(all_pops_bestfit.map_to_std[i]), sizeof(int),1,stt_fp) !=1) {
      fclose(stt_fp);
      return FALSE;
    }
  }
  for (i=0;i<All_Pops_Bestfit_Numfields;i++) {
    if(fread(&(all_pops_bestfit.field_ends[i]), sizeof(int),1,stt_fp) !=1) {
      fclose(stt_fp);
      return FALSE;
    }
  }
  
  if (sys_util_read_utility(all_pops_bestfit.utility, stt_fp) == FALSE) {
    fclose(stt_fp);
    return FALSE;
  }
  fclose(stt_fp);
  Unlock(restart_stt_filename);
	
  /* If was a restart and this process runs subpop 0,                    */
  /* and checkptfileprefix != restartfileprefix, create the              */
  /* file 'checkptfileprefix'99.stt from the information read above. */
  if (MustCopyRestartSttToCheckptStt == 1) {
    MustCopyRestartSttToCheckptStt = 0;
    CombineFileName(restart_stt_filename, checkptfileprefix, 99, ".stt");
    if (Lock(restart_stt_filename, 20) == FALSE) {
      fprintf(stderr, "\n*** ERROR. Couldn't lock %s in 20 tries"
	      , restart_stt_filename);
      exit(-1);
    }
    if ((stt_fp = fopen(restart_stt_filename, "wb")) == NULL) {
      fprintf(stderr,"\n**ERROR. Couldn't open %s.",restart_stt_filename);
      exit(-1);
    }

    /* Now write out updated performance stats to checkpt_stt_filename */
    if (fwrite(&(All_Pops_Bestfit_Popno), sizeof(int), 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fwrite(&(All_Pops_Bestfit_Chromsize), sizeof(int),1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fwrite(&(All_Pops_Bestfit_Numfields), sizeof(int),1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }

    if (fwrite(&all_pops_sumfitness, sizeof(double), 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fwrite(&all_pops_sum_best_fitness, sizeof(double),1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fwrite(&all_pops_online_denominator, sizeof(long),1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fwrite(&all_pops_best_fitness_count, sizeof(long),1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fwrite(&all_pops_neval, sizeof(long), 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fwrite(all_pops_bestfit.chrom
	       , All_Pops_Bestfit_Chromsize * sizeof(unsigned)
	       , 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fwrite(&(all_pops_bestfit.fitness), sizeof(double),1,stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fwrite(&(all_pops_bestfit.init_fitness)
	       , sizeof(double), 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fwrite(&(all_pops_bestfit.neval), sizeof(long), 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fwrite(&(all_pops_bestfit.generation), sizeof(int),1,stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    for (i=0;i<All_Pops_Bestfit_Numfields;i++) {
      if(fwrite(&(all_pops_bestfit.map_to_std[i])
		, sizeof(int),1,stt_fp) !=1) {
	fclose(stt_fp);
	return FALSE;
      }
    }
    for (i=0;i<All_Pops_Bestfit_Numfields;i++) {
      if(fwrite(&(all_pops_bestfit.field_ends[i])
		, sizeof(int),1,stt_fp) !=1) {
	fclose(stt_fp);
	return FALSE;
      }
    }
    
    if (sys_util_write_utility(all_pops_bestfit.utility, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    fclose(stt_fp);
    Unlock(restart_stt_filename);

  }
  return TRUE;
}


BOOL
UpdateCumStats()
/* Called at the beginning of EACH CYCLE (I.E., BEFORE FIRST POPULATION)  */
/* EXCEPT THE FIRST CYCLE OR THE FIRST CYCLE AFTER AN EXTEND from a saved */
/* checkpoint file.  NOT called before each subpop, but only once/cycle   */
/* through ALL subpops (UNLIKE globalstats, which is called after EACH    */
/* subpop is completed each cycle, in a new local best was found.         */
{

  int     i,j;
  FILE    *stt_fp;

  if (!single_process) {
    CombineFileName(restart_stt_filename, restartfileprefix, 99, ".stt");
    if (Lock(restart_stt_filename, 50) == FALSE) {
      fprintf(stderr, "\n*** OOPS! UpdateCumStats unable to lock %s."
	      , restart_stt_filename);
      fprintf(stderr, "\n   All later printing of cum. perf.");
      fprintf(stderr, "\n   statistics is probably in error.  BEWARE.\n");
      fprintf(outfp, "\n*** ERROR: UpdateCumStats unable to lock %s."
	      , restart_stt_filename);
      fprintf(outfp, "\n   All later printing of cum. perf.");
      fprintf(outfp, "\n   statistics is probably in error.  BEWARE.\n");
      return FALSE;
    }

    if ((stt_fp = fopen(restart_stt_filename, "r+b")) == NULL) {
      fprintf(stderr, "\n*** Locked but couldn't open %s.  Beware"
	      ,restart_stt_filename);
      Unlock(restart_stt_filename);
      return FALSE;
    }
    rewind(stt_fp);

    /* Find out length of chromosome, etc., for population whose statistics */
    /* are recorded as all_pops numbers here, so can malloc space for reading*/
    /* them in. */
    if (fread(&(All_Pops_Bestfit_Popno), sizeof(int), 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fread(&(All_Pops_Bestfit_Chromsize), sizeof(int), 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fread(&(All_Pops_Bestfit_Numfields), sizeof(int), 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }

    /* Get the variable-length fields in bestever structs malloc'd */
    bestever_variable_malloc();

    if (fread(&all_pops_sumfitness, sizeof(double), 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fread(&all_pops_sum_best_fitness, sizeof(double), 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fread(&all_pops_online_denominator, sizeof(long), 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fread(&all_pops_best_fitness_count, sizeof(long), 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fread(&all_pops_neval, sizeof(long), 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fread(all_pops_bestfit.chrom
	      , All_Pops_Bestfit_Chromsize * sizeof(unsigned)
	      , 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fread(&(all_pops_bestfit.fitness), sizeof(double), 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fread(&(all_pops_bestfit.init_fitness),sizeof(double),1,stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fread(&(all_pops_bestfit.neval), sizeof(long), 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fread(&(all_pops_bestfit.generation), sizeof(int), 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    for (i=0;i<All_Pops_Bestfit_Numfields;i++) {
      if(fread(&(all_pops_bestfit.map_to_std[i]), sizeof(int),1,stt_fp) !=1) {
	fclose(stt_fp);
	return FALSE;
      }
    }
    for (i=0;i<All_Pops_Bestfit_Numfields;i++) {
      if(fread(&(all_pops_bestfit.field_ends[i]), sizeof(int),1,stt_fp) !=1) {
	fclose(stt_fp);
	return FALSE;
      }
    }
    
    if (sys_util_read_utility(all_pops_bestfit.utility, stt_fp) == FALSE) {
      fclose(stt_fp);
      return FALSE;
    }

    /* Now, rewind the file in preparation for writing updated values to it. */
    rewind(stt_fp);
  }
  /* Now update these from current processor's just-completed cycle: */
  all_pops_sumfitness += local_cycle_sumfitness;
  all_pops_sum_best_fitness += local_cycle_sum_best_fitness;
  all_pops_online_denominator += local_cycle_online_denominator;
  all_pops_best_fitness_count += local_cycle_best_fitness_count;
  /* update all_pops_neval AFTER bestfit stuff uses current value */
  
  if (local_bestfit.init_fitness > all_pops_bestfit.init_fitness) {
    All_Pops_Bestfit_Popno = local_bestfit_popno;
    All_Pops_Bestfit_Chromsize = local_bestfit_chromsize;
    All_Pops_Bestfit_Numfields = local_bestfit_numfields;
    all_pops_bestfit.init_fitness = local_bestfit.init_fitness;
    all_pops_bestfit.fitness = local_bestfit.fitness;

    /* Now are writing new information from the new best individual */
    /* just found in this cycle (but NOT NECESSARILY in the */
    /* just-completed subpop) to the stt file. */
    for (i = 0; i < chromsize; i++)
      all_pops_bestfit.chrom[i] = local_bestfit.chrom[i];
    all_pops_bestfit.generation = local_bestfit.generation;
    all_pops_bestfit.neval = local_bestfit.neval;
    for (i=0;i<numfields;i++) {
      all_pops_bestfit.map_to_std[i] = local_bestfit.map_to_std[i];
      all_pops_bestfit.field_ends[i] = local_bestfit.field_ends[i];
    }
    sys_util_copy_utility(all_pops_bestfit.utility
		     , local_bestfit.utility);

    if(!single_process) {    
      if (quiet < 2) {
	fprintf(outfp, "\n *** Now at End of All Pops in Cycle %d ***\n"
		, cycle);
	fprintf(outfp, "\n*** Achieved New All_Pops Best This Cycle:");
	fprintf(outfp, " Its All_Pops_Neval= %ld ***\n"
		, all_pops_bestfit.neval);
	fprintf(outfp," raw fitness = %20.10f, gen = %d, chromosome =\n"
		, all_pops_bestfit.init_fitness
		, all_pops_bestfit.generation);
	writechrom(all_pops_bestfit.chrom);
	if (alpha_size != 2) writefields(all_pops_bestfit.chrom);
	if (using_inversion) {
	  fprintf(outfp,"\nIts inversion map_to_std:\n");
	  for (i=0;i<((numfields-1)/16) + 1;i++) {
	    for (j=0;j<16 && i * 16 + j < numfields;j++) {
	      fprintf(outfp, "%d "
		      , all_pops_bestfit.map_to_std[16*i+j]);
	    }
	    fprintf(outfp, "\n");
	  }
	}
      }
      app_new_global_best_report(); /*Do even if quiet == 3;user controls*/

      if (quiet < 2) {
	fprintf(outfp, "\nAt the END of the CYCLE in which the New ");
	fprintf(outfp, "All_Pops Bestfit was found,\n");
	fprintf(outfp
		, "On-Line Perf. = %20.10f, Off-Line Perf. = %20.10f\n"
		, all_pops_sumfitness / all_pops_online_denominator
		, all_pops_sum_best_fitness / all_pops_best_fitness_count);

      }
    }
  }
  
  all_pops_neval += local_cycle_neval;
  
  if (!single_process) {
    /* Now write out updated performance stats */
    if (fwrite(&(All_Pops_Bestfit_Popno), sizeof(int), 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fwrite(&(All_Pops_Bestfit_Chromsize), sizeof(int), 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fwrite(&(All_Pops_Bestfit_Numfields), sizeof(int), 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }

    if (fwrite(&all_pops_sumfitness, sizeof(double), 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fwrite(&all_pops_sum_best_fitness, sizeof(double), 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fwrite(&all_pops_online_denominator, sizeof(long), 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fwrite(&all_pops_best_fitness_count, sizeof(long), 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fwrite(&all_pops_neval, sizeof(long), 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fwrite(all_pops_bestfit.chrom
	       , All_Pops_Bestfit_Chromsize* sizeof(unsigned),1,stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fwrite(&(all_pops_bestfit.fitness), sizeof(double), 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fwrite(&(all_pops_bestfit.init_fitness),sizeof(double),1,stt_fp)!= 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fwrite(&(all_pops_bestfit.neval), sizeof(long), 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    if (fwrite(&(all_pops_bestfit.generation), sizeof(int), 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    for (i=0;i<All_Pops_Bestfit_Numfields;i++) {
      if(fwrite(&(all_pops_bestfit.map_to_std[i]), sizeof(int),1,stt_fp)!=1) {
	fclose(stt_fp);
	return FALSE;
      }
    }
    for (i=0;i<All_Pops_Bestfit_Numfields;i++) {
      if(fwrite(&(all_pops_bestfit.field_ends[i]), sizeof(int),1,stt_fp)!=1) {
	fclose(stt_fp);
	return FALSE;
      }
    }
    
    if (sys_util_write_utility(all_pops_bestfit.utility, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
    fclose(stt_fp);
    Unlock(restart_stt_filename);
  }
  return TRUE;
}


BOOL
InitializeCumStats()
/* Zero out the contents of the 'checkptfileprefix'99.stt file, which should */
/* be done only by subpop 0 at the beginning of the run, and          */
/* create the lock file to accompany it, if it doesn't exist.      */
{
  FILE    *stt_fp;
  char    byte_zero = '\0';
  int     int_zero = 0;
  unsigned uns_zero = 0;
  double  double_zero = 0.0;
  long    long_zero = 0L;
  int     i;
  unsigned  nbytes;


  /* First, create the companion lock file if it doesn't exist already. */
  if (!strlen(checkptfileprefix))
    strcpy(buffer, "sgackp");
  else
    strcpy(buffer, checkptfileprefix);
  CombineFileName(checkpt_stt_filename, buffer, 99, ".stt");
  if ((stt_fp = fopen(checkpt_stt_filename, "rb")) == NULL) {
    if(MakeLockable(checkpt_stt_filename) == FALSE) {
      fprintf(stderr, "\n*** OOPS, couldn't make %s lockable."
	      , checkpt_stt_filename);
      exit(-1);
    }
  }
  else
    fclose(stt_fp);

  if(Lock(checkpt_stt_filename, 50) == FALSE) {
    fprintf(stderr, "\n***Couldn't lock %s after made lockfile."
	    , checkpt_stt_filename);
    return FALSE;
  }

  if ((stt_fp = fopen(checkpt_stt_filename, "wb")) == NULL) {
    fprintf(stderr, "\n *** Initialization of %s failed."
	    , checkpt_stt_filename);
    exit (-1);
  }
  /* Initialize the newly created 'checkptfileprefix'99.stt file */
  if (fwrite(&popno, sizeof(int),1,stt_fp) != 1) { /* this popno */
    fclose(stt_fp);
    return FALSE;
  }
  if (fwrite(&chromsize, sizeof(int), 1, stt_fp) != 1) { /* this chromsize */
    fclose(stt_fp);
    return FALSE;
  }
  if (fwrite(&numfields, sizeof(int), 1,stt_fp) != 1) { /* dummy numfields */
    fclose(stt_fp);
    return FALSE;
  }
  if (fwrite(&double_zero, sizeof(double), 1, stt_fp) != 1) {
    fclose(stt_fp);
    return FALSE;
  }
  if (fwrite(&double_zero, sizeof(double), 1, stt_fp) != 1) {
    fclose(stt_fp);
    return FALSE;
  }
  if (fwrite(&long_zero, sizeof(long), 1, stt_fp) != 1) {
    fclose(stt_fp);
    return FALSE;
  }
  if (fwrite(&long_zero, sizeof(long), 1, stt_fp) != 1) {
    fclose(stt_fp);
    return FALSE;
  }
  if (fwrite(&long_zero, sizeof(long), 1, stt_fp) != 1) {
    fclose(stt_fp);
    return FALSE;
  }
  for (i=0;i<chromsize;i++) {   /* Using chromsize from pop 0 */
    if (fwrite(&uns_zero, sizeof(unsigned),1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
	
    }
  }
  if (fwrite(&double_zero, sizeof(double), 1, stt_fp) != 1) {
    fclose(stt_fp);
    return FALSE;
  }
  if (fwrite(&double_zero, sizeof(double), 1, stt_fp) != 1) {
    fclose(stt_fp);
    return FALSE;
  }
  if (fwrite(&long_zero, sizeof(long), 1, stt_fp) != 1) {
    fclose(stt_fp);
    return FALSE;
  }
  if (fwrite(&int_zero, sizeof(int), 1, stt_fp) != 1) {
    fclose(stt_fp);
    return FALSE;
  }
  for (i=0;i<numfields;i++) {   /* Using numfields from pop 0 */
    if (fwrite(&int_zero, sizeof(int), 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
  }
  for (i=0;i<numfields;i++) {   /* Using numfields from pop 0 */
    if (fwrite(&int_zero, sizeof(int), 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
  }
  
  nbytes = GetUtilitySize(); 
  for (i=0;i<nbytes;i++) {
    if (fwrite(&byte_zero, 1, 1, stt_fp) != 1) {
      fclose(stt_fp);
      return FALSE;
    }
  }

  fclose(stt_fp);
  Unlock(checkpt_stt_filename);
  return TRUE;
}

