/*   

  File manager file_state.c

  Contains routines to manage asynchronous file writing and reading by 
  multiple processes running on multiple processers.

  Notes for properly using the routines provided here:
 
    Routines that do writes call start_write() and use the filepointer
    returned to write data in any format, calling end_write() complete
    process.  The newly written data will not be available to any readers
    of the file, including the writing process until read_update() has been
    completed.

    Routines need to provide char *filename and, uninitialized, 
    FILE *fp and long states[2] for communication from start_write() to
    end_write().  The int retries is the number of times the writer will try
    to open a file before giving up.  On UNIX-NFS systems, no retries are
    required.
 
    A written file must be updated by a call to read_update() before a 
    subsequent write to prevent buffer being read while write procedes.  While 
    not updated, the most recently written buffer is not readable.  

    Multi-process mode:
    In the case of a read/write collision, start_write() will pause, then 
    clear the appropriate read flag and continue.  The reading process will 
    see the file has been overwritten at the end of the read.  

    Routines that do reads call start_read() and the use filepointer returned
    to read data in known format, calling end_read() to complete.
    Routines need to provide char *filename and int own_file, and
    uninitialized, FILE *fp, static unsigned long bit_flag and index, 
    int position and BOOL read_thru.  The input value, own_file, should be 1 if 
    the file is written to by this process and 0 if not. 

    If end_read() finds the buffer has been overwritten while the read was in 
    progress, it will return 0 and read_thru FALSE.
    Calling code may decide to re-read file.

  routines: 
      **** write ****
      FILE *start_write(char *filename, long states[2], int retries)
	  return of 0 means file cannot be opened, non-recoverable error
	  Checks states by reading the file.  In multi-process mode, Reads 
	  flags to determine if a read by another process is in progress, then
	  waits to see if flag clears, overwriting after WRITE_WAIT_LOOP tries. 
	  Sets file offset to write position.

	  start_write() calls 
	  int file_state_check(char* filename, long file_states[2], 
			       unsigned long flags[2]) 
	      return of 0 means file cannot be opened, non-recoverable error
	      Reads write and read states and read flags if file exists, 
	      otherwise, creates file and inits states and flags.
 
      int end_write(car *filename, FILE *fp, long *states, int retries)        
	  return of 0 means file could not be written or closed, nonrecoverable
	  Updates read and write states and closes file.

      int update_rstate(char *filename, int retries)
	  return of 0 means file could not be written or closed, nonrecoverable
	  Updates read state and closes file.
	  Must be called between file writes.

      **** read ****
      FILE *start_read(char *filename, int own_file, retries, 
		       unsigned long *bit_flag, int *index, *position)
	  return of 0 means file cannot be read now
	  Calculates and sets own read flag in file if multi-process.
	  Sets file offset to read position.

      int end_read(char *filename, FILE *fp, int own_file, retries,  
		   unsigned long bit_flag, iint index, position, BOOL *readthru)
	  return of 0 means file could not be closed, nonrecoverable, or that 
	  read needs to be repeated because of overwrite (readthru == FALSE) 
	  Checks and clears own read flag in file if multi-process, closes file.

  Additional notes:

    Considerable complication in the routines has resulted from the effort to
    keep one version that has been tested to run under Solaris and HPUX, and on
    DOS->PATHWORKS->UNIX.  The attempt to create a system that will guarantee to
    protect the integrity of data read from a file written to asynchronously by
    another process running on another processor, has not been completely
    successful, but has run without error in small-scale, heavy collision-
    producing situations.  These routines do not, as currently tested, protect
    files WRITTEN to by more than one processor, only those read by multiple
    processes, written to only by one.
    There are complicating factors in the UNIX-style buffered file IO process
    which we have dealt with by re-writing until file reads as we require.  This
    may appear unnecessary, but be aware that any changes to these checking 
    loops result in errors when multiple processes are reading and writing the 
    same files.  These checking loops are found in end_write(), read_update(),
    start_read() and end_read().  
    Because some systems return a NULL pointer on attempts to open a file being 
    read by another process, loops external to the checking loops, and limited 
    by the input variable, retries, allow repeated fopen() attempts.
    Changing calls from the fopen(), fread(), etc, family to the open(), read(),
    etc. family, did not alleviate any problems, but resulted in a slight
    degradation of performance.


  file format:
  
  position | use                    | size/type | file offset pointer
  -----------------------------------------------------------------------
      0    | read state             | 1/long    | 0
      1    | write state            | 1/long    | sizeof(long)
      2    | buffer 0 position      | 1/long    | file_pos[0] = 2*sizeof(long)
      3    | buffer 1 position      | 1/long    | file_pos[1] = 3*sizeof(long)
      4    | read flag for buffer 0 | FLG_LEN   | file_pos[2] = 4*sizeof(long)
	   |                        |  /long    |
  4+FLG_LEN| read flag for buffer 1 | FLG_LEN   | file_pos[3] 
	   |                        |  /long    |   = (4+FLG_LEN*sizeof(long)
  -----------------------------------------------------------------------
     12    | buffer 0               |  x        | file_pos[4] = .
     ...   |                        |           | buffer 0 position
     ...   |                        |           |
     ...   |                        |           |
  -----------------------------------------------------------------------
  12 + len | buffer 1               |  x        | buffer 1 position
     ...   |                        |           |
     ...   |                        |           |
 12+2*len-1| end of file            |           |
  -----------------------------------------------------------------------


  state table:

   write | read  | write       | read
   state | state | position ptr| position ptr
  -------------------------------------------
     0   |   x   | file_pos[0] | no read
	 |       |             |
     1   |   0   | file_pos[1] | no read
     1   |   1   | file_pos[1] | file_pos[0]
	 |       |             |
  %2 = 0 |   0   | file_pos[0] | file_pos[0]
  %2 = 0 |   1   | file_pos[0] | file_pos[1]
	 |       |             |
  %2 = 1 |   0   | file_pos[1] | file_pos[1]
  %2 = 1 |   1   | file_pos[1] | file_pos[0]


  Written 1995 by Leslie Thomas Wall Hoppensteadt
  Copyright 1995 by Michigan State University Board of Trustees
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#ifndef __MWERKS__
  #include <sys/types.h>
  #include <sys/stat.h>
#else
  #include <types.h>
  #include <stat.h>
#endif
#include <fcntl.h>
#include <math.h>
#include "external.h"
#ifndef  HIBITL
#define  HIBITL 0x80000000
#endif
#ifndef  MAXLONG
#define MAXLONG ((long)(~HIBITL))
#endif

/* Added for compatibility with some systems where found in a strange */
/* place, 5/31/96, EDG. */
#ifndef SEEK_SET
#define SEEK_SET 0
#endif

int TotalFileOpenFailures = 0;  /* put in sga.h */

#define WRITE_WAIT_LOOP 5
#define FAIL_COUNT 50
#define FLG_LEN 4
#ifndef TRUE  /* declarations for borland c */
#define TRUE 1
#define FALSE 0
/*#define DEBUG_FILES2 */
FILE *fopen();  
#endif



static int write_wait = 0;
static long file_pos[5] = {  2*sizeof(long), 3*sizeof(long), 4*sizeof(long),
			       (4+FLG_LEN)*sizeof(long), 
			       (4+FLG_LEN*2)*sizeof(long)};

/* first time for this file */
int init_write(fname)
     char* fname;
{
  FILE *fp;
  int i;
  long states[2];
  unsigned long flags[2*FLG_LEN];
  
  states[0] = 0;
  states[1] = 0;      
  for (i = 0; i < 2*FLG_LEN; i++)
    flags[i] = 0;
  
  if ((fp = fopen(fname, "wb"))) {
    if (fwrite(states, sizeof(long), 2, fp)) 
      if (fwrite(&(file_pos[4]), sizeof(long), 1, fp)) /* set 0 position */
	if (fwrite(&(file_pos[4]), sizeof(long), 1, fp)) /* dummy */
	  if (fwrite(flags, sizeof(long), 2*FLG_LEN, fp))  /* zero flags */ 
	    return (fclose(fp) != EOF);
  
    if (fp)
      fclose(fp);
    return 0;
  }

  return 1;
}

FILE *start_write(fname, states, retries)
     char* fname;
     long *states;
     int retries;
{
  FILE* fp;
  unsigned long mask, flags[2*FLG_LEN], *new_ptr, *old_ptr, *flag_ptr;
  long pos;
  int i, check, reader, buf, again;
  int error = 0;
  int spin();

  if (retries < 1)
    retries = 1;

  again = 0;
  while (!(file_state_check(fname, states, flags)) && (again < retries)) 
    again++;

  if (again) {
    TotalFileOpenFailures += again;
    if (retries <= again)
      return 0;
  }

  if (single_process == FALSE) {
    buf = (states[1]%2);
    old_ptr = &flags[buf*FLG_LEN];
    new_ptr = &flags[(!buf)*FLG_LEN];
    check = 0;
    for (i = 0; i < FLG_LEN; i++) 
      if (!(new_ptr[i] & old_ptr[i]))
	check += old_ptr[i];

    again = 0;
    while (check && again < retries) {
      /* do something to let other process flush file output */
      spin(check);
      check = 0;
      if (file_state_check(fname, states, flags)) {
	buf = (states[1]%2);
	old_ptr = &flags[buf*FLG_LEN];
	new_ptr = &flags[(!buf)*FLG_LEN];
	for (i = 0; i < FLG_LEN; i++) 
	  if (!(new_ptr[i] & old_ptr[i]))
	    check += old_ptr[i];

	if (WRITE_WAIT_LOOP < ++error) 
	  break;
      }
      else
	again++;
    }
    
    if (check) {
#ifdef DEBUG_FILES
      unsigned long test;

      flag_ptr = &flags[(states[1]%2)*FLG_LEN];
      reader = 0;
      mask = 1;
      test = *flag_ptr;
      printf("Writing %s, collision with reading process(es):", fname);
      for (i = 0; i < FLG_LEN; i++) {
	while (test) {
	  if (test & mask)
	    printf(" %d", reader);
	  test = test >> 1;
	  reader++;
	}
	test = *(++flag_ptr);
      }
    
      printf(". Buffer %d overwritten by %d.\n", (states[1]%2), startpopnum);
      printf("states %ld %ld; flags", states[0], states[1]);
      for (i = 0; i < FLG_LEN*2; i++)
	if (0 <= flags[i])
	  printf(" %ld", flags[i]);
	else
	  printf(" MAX%ld", flags[i]);
#endif

      for (i = 0; i < FLG_LEN; i++)
	flags[(states[1]%2)*FLG_LEN + i] = 0;

#ifdef DEBUG_FILES  
      printf(" ->");
      for (i = 0; i < FLG_LEN*2; i++)
	if (0 <= flags[i])
	  printf(" %ld", flags[i]);
	else
	  printf(" MAX%ld", flags[i]);
      printf("\n\n");
#endif

      again = 0;
      while (!(fp = fopen(fname, "r+b")) && again++ < retries)
	;
      if (again) {
	TotalFileOpenFailures += again;
	if (!fp)
	  return 0;
      }

      fseek(fp, file_pos[2], SEEK_SET);
      fwrite(flags, sizeof(long), 2*FLG_LEN, fp);
      fclose(fp);
    } /* end if (check) */
  } /* end if (!single_porcess) */

  again = 0;
  while (!(fp = fopen(fname, "r+b")) && again < retries)
    again++;
  if (again) {
    TotalFileOpenFailures += again;
    if (!fp)
      return 0;
  }
  if (!(fseek(fp, file_pos[states[1]%2], SEEK_SET)))
    if (fread(&pos, sizeof(long), 1, fp))
      if (!(fseek(fp, pos, SEEK_SET)))
	return fp;
  
  if (fp) 
    fclose(fp);
  return 0;
}

int end_write(fname, fp, states, retries)
     char *fname;
     FILE *fp;
     long *states;
{
  long instates[2];
  int looper, again;

  states[0] = 0;    /* no read update */

  if (states[1]) {
    states[1] = (++states[1])%MAXLONG;    
    rewind(fp);
    if (fwrite(states, sizeof(long), 2, fp)) 
      if (single_process == TRUE)
	return (fclose(fp) != EOF);

    fflush(fp);
    fclose(fp);

    if (retries < 1)
      retries = 1;
    looper = again = 0;
    while (!(fp = fopen(fname, "r+b")) && again++ < retries) 
      ;
    if (again) {
      TotalFileOpenFailures += again;
      if (!fp)
	return 0;
    }

    fread(instates, sizeof(long), 2, fp);
    while (instates[0] != states[0] || instates[1] != states[1]
	   && looper < FAIL_COUNT) {
      rewind(fp);
      fwrite(states, sizeof(long), 2, fp);
      fclose(fp);
      looper++;
      fp = fopen(fname, "r+b");
      fread(instates, sizeof(long), 2, fp);
    }
#ifdef DEBUG_FILES2
    if (looper)
      printf("%d loop(s) to write STATE UPDATE.\n", looper);
#endif
    if (FAIL_COUNT <= looper) 
      fprintf(stderr, 
	      "Unable to verify correct file write of %s after %d attempts!!\n",
	      fname, looper);

    return (fclose(fp) != EOF);  
  }

  /* first time write, set start position for next write */
  else {
    long pos = ftell(fp);       /* end of 0 buffer */
    states[1] = 1;
    rewind(fp);

    if (fwrite(states, sizeof(long), 2, fp))    
      if (!(fseek(fp, file_pos[1], SEEK_SET))) {
	fwrite(&pos, sizeof(long), 1, fp);
	if (single_process == TRUE)
	  return (fclose(fp) != EOF);
	fclose(fp);

	looper = again = 0;
	while (!(fp = fopen(fname, "r+b")) && again++ < retries) 
	  ;
	if (again) {
	  TotalFileOpenFailures += again;
	  if (!fp)
	    return 0;
	}
	fread(instates, sizeof(long), 2, fp);
	while (instates[0] != states[0] || instates[1] != states[1]
	       && looper < FAIL_COUNT) {
	  rewind(fp);
	  fwrite(states, sizeof(long), 2, fp);
	  fclose(fp);
	  looper++;
	  while (!(fp = fopen(fname, "r+b")) && again++ < retries)
	    ;
	  if (again) {
	    TotalFileOpenFailures += again;
	    if (!fp)
	      return 0;
	  }
	  fread(instates, sizeof(long), 2, fp);
	}
#ifdef DEBUG_FILES2
	if (looper)
	  printf("%d loop(s) to write STATE UPDATE.\n", looper);
#endif
	if (FAIL_COUNT <= looper) {
	  fprintf(stderr, "Unable to verify correct file write of %s after %d",
	      fname, looper);
	  fprintf(stderr, " attempts!!\n");
	}

	return (fclose(fp) != EOF); 
      }
  }
  
  if (fp)
    fclose(fp);
  return 0;
}

int update_rstate(fname, retries)
     char *fname;
     int retries;
{
  FILE *fp = 0;
  int looper, again;
  long update_done = 1;

  if (retries < 1)
    retries = 1;
  looper = again = 0;

  while (!(fp = fopen(fname, "r+b")) && again++ < retries)
    ;
  if (again) {
    TotalFileOpenFailures += again;
    if (!fp)
      return 0;
  }

  if (fwrite(&update_done, sizeof(long), 1, fp)) {
    if (single_process)
      return (fclose(fp) != EOF);

    fflush(fp); /* why not? */
    fclose(fp);
    
    while (!(fp = fopen(fname, "r+b")) && again++ < retries)
      ;
    if (again) {
      TotalFileOpenFailures += again;
      if (!fp)
	return 0;
    }
    fread(&update_done, sizeof(long), 1, fp);
    while (!update_done && looper < FAIL_COUNT) {
      fclose(fp);
      update_done = 1;
      looper++;
      while (!(fp = fopen(fname, "r+b")) && again++ < retries)
	;
      if (again) {
	TotalFileOpenFailures += again;
	if (!fp)
	  return 0;
      }
      fwrite(&update_done, sizeof(long), 1, fp);
      fclose(fp);
      update_done = 0;
      while (!(fp = fopen(fname, "r+b")) && again++ < retries)
	;
      if (again) {
	TotalFileOpenFailures += again;
	if (!fp)
	  return 0;
      }
      fread(&update_done, sizeof(long), 1, fp);
    }
#ifdef DEBUG_FILES2
    if (looper)
      printf("%d loop(s) to write READ UPDATE.\n", looper);
#endif
    if (FAIL_COUNT <= looper) {
      fprintf(stderr, "Unable to verify correct file write of %s after %d",
	      fname, looper);
      fprintf(stderr, " attempts!!\n");
    }

    return (fclose(fp) != EOF);
  }

  if (fp)
    fclose(fp);
  return 0;
}

int file_state_check(fname, states, flags)
     char* fname;
     long *states;
     unsigned long *flags;
{
  FILE *fp;
  int i;
  int check = 0;

  if ((fp = fopen(fname, "rb"))) 
    if ((fread(states, sizeof(long), 2, fp)))
      if (fread(flags, sizeof(long), 2, fp)) /* dummy read */
	if (fread(flags, sizeof(long), 2*FLG_LEN, fp))   /* read flags */
	  return (fclose(fp) != EOF);

  /* first time for this file */
  if ((fp = fopen(fname, "wb"))) {
    /* no read update done,  write state 0 */
    states[0] = 0;
    states[1] = 0;      
    for (i = 0; i < 2*FLG_LEN; i++)
      flags[i] = 0;
    if (fwrite(states, sizeof(long), 2, fp)) 
      if (fwrite(&(file_pos[4]), sizeof(long), 1, fp)) /* set 0 position */
	if (fwrite(&(file_pos[4]), sizeof(long), 1, fp)) /* dummy */
	  if (fwrite(flags, sizeof(long), 2*FLG_LEN, fp))  /* zero flags */ 
	    return (fclose(fp) != EOF);
  }

  if (fp)
    fclose(fp);
  return 0;
}


FILE *start_read(fname, own_file, retries, mask, index, buf_pos)
     char* fname;
     int own_file, retries;
     unsigned long *mask, *index;
     int *buf_pos;
{
  FILE* fp;
  long states[2], pos;
  unsigned long flags[2*FLG_LEN];
  int flag_pos, looper, again;

  if (retries < 1)
    retries = 1;
  again = 0;

  if (single_process || own_file) {
    while (!(fp = fopen(fname, "rb")) && again++ < retries)
      ;
    if (again) {
      TotalFileOpenFailures += again;
      if (!fp)
	return 0;
    }

    if (fread(states, sizeof(long), 2, fp)) 
      /* if read update performed or both buffers written */
      if (states[0] || 1 < states[1]) {
	*buf_pos = (int)(((states[0]+states[1])%2));
	if (!(fseek(fp, file_pos[*buf_pos], SEEK_SET))) 
	  if (fread(&pos, sizeof(long), 1, fp)) 
	    if (!(fseek(fp, pos, SEEK_SET)))
	      return fp;
      }
    
    if (fp)
      fclose(fp);
    return 0;
  }

  again = 0;
  while (!(fp = fopen(fname, "r+b")) && again++ < retries)
     ;
  if (again) {
    TotalFileOpenFailures += again;
    if (!fp)
      return 0;
  }

  if (fread(states, sizeof(long), 2, fp)) 
    /* if read update performed or both buffers written */
    if (states[0] || 1 < states[1]) {
      *buf_pos = (int)(((states[0]+states[1])%2));
      flag_pos = *buf_pos*FLG_LEN;
      /* set flag to show we are reading this buffer */
      if (!(fseek(fp, file_pos[2], SEEK_SET))) 
	if (fread(flags, sizeof(long), 2*FLG_LEN, fp)) {
	  get_bitflag(&(flags[flag_pos]), mask, index);
	  flags[flag_pos+*index] |= *mask;
	  fseek(fp, file_pos[2], SEEK_SET);
	  fwrite(flags, sizeof(long), 2*FLG_LEN, fp);
	  fflush(fp);
	  fclose(fp);
	}
	
      /* check to see if write completed */
      looper = again = 0;
      while (!(fp = fopen(fname, "r+b")) && again++ < retries) 
	;
      if (again) {
	TotalFileOpenFailures += again;
	if (!fp)
	  return 0;
      }

      if (!fseek(fp, file_pos[2], SEEK_SET)) {
	fread(flags, sizeof(long), 2*FLG_LEN, fp);
	while (!(flags[flag_pos+*index] & *mask) && looper < FAIL_COUNT) {
	  flags[flag_pos+*index] |= *mask;
	  fseek(fp, file_pos[2], SEEK_SET);
	  fwrite(flags, sizeof(long), 2*FLG_LEN, fp);
	  fflush(fp);
	  fclose(fp);
	  looper++;

	  while (!(fp = fopen(fname, "r+b")) && again++ < retries) 
	    ;
	  if (again) {
	    TotalFileOpenFailures += again;
	    if (!fp)
	      return 0;
	  }

	  fseek(fp, file_pos[2], SEEK_SET);
	  fread(flags, sizeof(long), 2*FLG_LEN, fp);
	}

	if (FAIL_COUNT <= looper) {
	  fprintf(stderr, "Unable to verify correct file read of %s after %d",
		  fname, looper);
	  fprintf(stderr, " attempts!!\n");
	}

	fclose(fp);
#ifdef DEBUG_FILES2
	if (looper)
	  printf("%d loop(s) to SET read bitflag %ld.\n",looper, *mask);
#endif
#ifdef DEBUG_FILES3
	  printf("Set bitflag %ld, %d by %d.\n", *mask, *index,startpopnum);
#endif
      }

      /* open for reading data */
      again = 0;
      while (!(fp = fopen(fname, "rb")) && again++ < retries) 
	;
      if (again) {
	TotalFileOpenFailures += again;
	if (!fp)
	  return 0;
      }

      if (!(fseek(fp, file_pos[*buf_pos], SEEK_SET))) 
	if (fread(&pos, sizeof(long), 1, fp)) 
	  if (!(fseek(fp, pos, SEEK_SET)))
	    return fp;
    }
  
  if (fp)
    fclose(fp);
  return 0;
}

int end_read(fname, fp, own_file, retries, mask, index, buf_pos, readthru)
     char *fname;
     FILE *fp;
     int own_file, retries;
     unsigned long mask, index;
     int buf_pos;
     BOOL *readthru;
{
  unsigned long flags[2*FLG_LEN];
  int ndx;
  int looper, again;

  if (single_process == TRUE || own_file) {
    *readthru = TRUE;
    return (fclose(fp) != EOF);
  }  
  
  fclose(fp);
  ndx = buf_pos + (int)index;
  looper = again = 0;
  while (!(fp = fopen(fname, "r+b")) && again++ < retries) 
    ;
  if (again) {
    TotalFileOpenFailures += again;
    if (!fp)
      return 0;
  }

  if (!(fseek(fp, file_pos[2], SEEK_SET)))
    if (fread(flags, sizeof(long), 2*FLG_LEN, fp)) {
      *readthru = (flags[ndx] & mask) ? TRUE: FALSE;
      /* we need to clear our read flag, even if it looks clear */
      flags[ndx] = flags[ndx] & ~(mask);
      fseek(fp, file_pos[2], SEEK_SET);
      fwrite(flags, sizeof(long), 2*FLG_LEN, fp);
      fflush(fp);
      fclose(fp);

      /* open to see if write completed */
      again = 0;
      while (!(fp = fopen(fname, "r+b")) && again++ < retries) 
	;
      if (again) {
	TotalFileOpenFailures += again;
	if (!fp)
	  return 0;
      }

      fseek(fp, file_pos[2], SEEK_SET);
      fread(flags, sizeof(long), 2*FLG_LEN, fp);
      while ((flags[ndx] & mask) && looper < FAIL_COUNT) {
	flags[ndx] = flags[ndx] & ~(mask);
	fseek(fp, file_pos[2], SEEK_SET);
	fwrite(flags, sizeof(long), 2*FLG_LEN, fp);
	fclose(fp);
	looper++;
	while (!(fp = fopen(fname, "r+b")) && again++ < retries) 
	  ;
	if (again) {
	  TotalFileOpenFailures += again;
	  if (!fp)
	    return 0;
	}
	fseek(fp, file_pos[2], SEEK_SET);
	fread(flags, sizeof(long), 2*FLG_LEN, fp);
      }
#ifdef DEBUG_FILES2
      if (looper) 
	printf("%d loop(s) to CLEAR read bitflag %ld, %ld.\n",
	       looper,mask,index);
#endif
#ifdef DEBUG_FILES3
      printf("Cleared bitflag %ld, %d by %d.\n\n", mask,index,startpopnum);
#endif
      if (FAIL_COUNT <= looper) {
	fprintf(stderr, "Unable to verify correct file read of %s after %d",
		fname, looper);
	fprintf(stderr, " attempts!!\n");
      }

      return ((fclose(fp) != EOF) && (*readthru == TRUE));  
    }
}


void get_bitflag(flags, mask, index)
     unsigned long *flags, *mask, *index;
{
  unsigned long bitstart, msk;
  int loopstart, ndx;
  int debug = 0;

  ndx = startpopnum/UINTSIZE;
  loopstart = ndx;
  msk = 1;
  msk = msk << (startpopnum%UINTSIZE);
  bitstart = msk;

  while ((flags[ndx] & msk)) {
#ifdef DEBUG_FILES4
    if (!debug) {
      printf("++ for %ld, %ld", msk, ndx);
      debug++;
    }
#endif
    if (msk < HIBITL) 
      msk = msk << 1;
    else {
      ndx = (++ndx)%FLG_LEN;
      msk = 1;
    }
#ifdef DEBUG_FILES4
  if (debug) 
    printf(" to %ld, %ld", msk, ndx);
#endif
    if (ndx == loopstart) 
      if (msk == bitstart)
	break;
  }     

  *index = ndx;
  *mask = msk;

#ifdef DEBUG_FILES4
  if (debug) {
    printf(". flag");
    for (debug = 0; debug < FLG_LEN; debug++)
      if (0 <= flags[debug])
	printf(" %ld", flags[debug]);
      else
	printf(" MAX%ld", flags[debug]);
    printf(".\n");
  }
#endif
}

int spin(seed)
     int seed;
{
  double slow;

  slow = fabs((double)seed) + 1.1;
  slow = sqrt(slow);
  slow = acos(fmod(slow, 1.0));

  return (int)slow;
}

#ifdef DEBUG_FILES    
int file_test(inname, outfptr, known_len)
     char *inname;
     FILE *outfptr;
     int known_len;
{
  long input[1000];
  int i, j, k, header_len;

  FILE *fp = fopen(inname, "rb");
  if (!fp) {
    printf("file %s not found.\n", inname);
    return 0;
  }

  for (i = 0; i < 1000; i++)
    input[i] = 0;

  i = 0;
  header_len = 4 + FLG_LEN*2;
  while ((i < header_len) && (fread(&(input[i]), sizeof(long), 1, fp)))
    i++;

  fprintf(outfptr, "read state %ld; write state %ld; ", input[0], input[1]); 
  fprintf(outfptr, "buffers %ld %ld; flags ", input[2], input[3]);
  for (j = 4; j < header_len; j++)
    if (0 <= input[j])
      fprintf(outfptr, "%ld ", input[j]);
    else
      fprintf(outfptr, "MAX%ld ", input[j]);
  fprintf(outfptr, "\n");

  if (-1 < known_len) {
    while ((i < 1000) && (fread(&(input[i]), sizeof(int), 1, fp)))
      i++;

    fprintf(outfptr, "first data buffer as %d ints:\n", known_len);
    for (k = j; k < i; k++) {
      fprintf(outfptr, "%ld ", input[k]);
      if (k && k == (known_len+j-1))
	fprintf(outfptr, "\nsecond data buffer:\n");
    }
    fprintf(outfptr, "\n");
  }

  fclose(fp);
  return i;
}
#endif
