/*---------------------------------------------------------------------------*/
/* invercla.c - perform a "classical" inversion on the mapping of the        */
/*              chromosome in a subpopulation to the "standard" rep used     */
/*              in the objective function.  Note that it is harder,          */
/*              probabilistically speaking, to rearrange loci near the ends  */
/*              of the chromosome, so the ends have, effectively, tighter    */
/*              linkage among loci of a given separation than do the loci    */
/*              in the middle of the chromosome.  This may, depending on the */
/*              representation, act as an advantage or a disadvantage, when  */
/*              compared with a "circular" inversion operator, which treats  */
/*              the chromosome as a ring, much as two-point crossover does.  */
/*              Modified to handle fields of different lengths, Jun, 1996,   */
/*              by Erik Goodman. */

#include <math.h>
#include "external.h"

int *new_field_sizes = NULL;
int *new_field_ends = NULL;
int *old_map_to_std = NULL;
int *inverse_new_map_to_std = NULL;
int *old_field_ends = NULL;
unsigned int *tempchrom = NULL;


char which_inversion[] = "invercla.c";

void
inversion(map)
/* Perform an inversion on this subpopulation's mapping of fields to */
/* the "standard" order used in the objective function.              */
int *map;
{

  unsigned nbytes;
  int inversion_start, inversion_end, tmp, i, j, k;
  int prev_end, fldlen, fldsiz, start, end;

  /* Pick two points on chromosome for ends of segment to invert */
  inversion_start = rnd(0, numfields-1);
  do {
      inversion_end = rnd(0, numfields-1);
  } while (inversion_end == inversion_start);

  if (inversion_start > inversion_end) {
      tmp = inversion_start;
      inversion_start = inversion_end;
      inversion_end = tmp;
  }

  /* If fields are of different sizes, must save old map, too. */
  if (alpha_size < 2) {
      if (old_map_to_std == NULL) {
	nbytes = numfields * sizeof(int);
	if ((old_map_to_std = (int *) malloc(nbytes)) == NULL)
	  nomemory("old_map_to_std");
      }
      if (old_field_ends == NULL) {
	nbytes = numfields * sizeof(int);
	if ((old_field_ends = (int *) malloc(nbytes)) == NULL)
	  nomemory("old_field_ends");
      }
 
      for (i=0;i<numfields;i++) {
	old_map_to_std[i] = map[i];
	old_field_ends[i] = field_ends[i];
      }
    }

    /* Alter the map, almost "in place".  Note integer division. */
    for (i=0; i<(inversion_end - inversion_start+1)/2;i++) {
	tmp = map[inversion_start + i];
	map[inversion_start + i] = map[inversion_end - i];
	map[inversion_end - i] = tmp;
    }
    /* Now, for each individual in this (sub)population, make the same */
    /* alteration in its chromosome as just made on the map, to keep the */
    /* chromosomes MATCHING the map. Easy if all fields same size, next. */
    if (alpha_size >=2) {
      /* ALL FIELDS ARE OF SAME SIZE -- CAN JUST REMAP ALMOST IN PLACE */
      for (j=0;j<popsize;j++) {
	for (i=0; i<(inversion_end - inversion_start+1)/2;i++) {
	    tmp = getfield(inversion_start+i, numfields, oldpop[j].chrom);
	    putfield(inversion_start+i
		     , numfields
		     , getfield(inversion_end-i, numfields, oldpop[j].chrom)
		     , oldpop[j].chrom);
	    putfield(inversion_end-i, numfields, tmp, oldpop[j].chrom);
	}
      }
  }
  else {
    /* MUST REARRANGE FIELDS IN POPULATION AND FIELDS ARE OF DIFFERENT SIZES.*/
    /* Do it into temporary chrom, then copy back -- note */
    /* that can't use the usual "getfield" and "putfield" with the temporary */
    /* chrom, as they use field_ends, which depends on field_sizes, which */
    /* changes under the inversion being done, if diff. length fields are in */
    /* use.  */
 
    /*  0.  malloc temp memory needed for transition to new rep. */
    if (new_field_sizes == NULL) {
      nbytes = numfields * sizeof(int);
      if ((new_field_sizes = (int *) malloc(nbytes)) == NULL)
	nomemory("new_field_sizes");
    }
    /* allocate memory for inverse of new_map_to_std */
    if (inverse_new_map_to_std == NULL) {
      nbytes = numfields * sizeof(int);
      if ((inverse_new_map_to_std = (int *) malloc(nbytes)) == NULL)
	nomemory("inverse_new_map_to_std");
    }
    /* allocate memory for new_field_ends */
    if (new_field_ends == NULL) {
      nbytes = numfields * sizeof(int);
      if ((new_field_ends = (int *) malloc(nbytes)) == NULL)
	nomemory("new_field_ends");
    }
    /* malloc the temporary chromosome needed below */
    if (tempchrom == NULL) {
      nbytes = chromsize * sizeof(unsigned);
      if ((tempchrom = (unsigned *) malloc(nbytes)) == NULL)
	nomemory("tempchrom");
    }
    /*  1. invert field_sizes into new_field_sizes */
    for (i=0;i<numfields;i++)
      inverse_new_map_to_std[map[i]] = i;
    for (i=0;i<numfields;i++)
      new_field_sizes[inverse_new_map_to_std[old_map_to_std[i]]]
	 = field_sizes[i];
    /* new_field_sizes now contains proper values for re-arranged chromosome. */
    /*  2. calculate new_field_ends to replace field_ends for this subpop. */
    prev_end = 0;
    for (i=0;i<numfields;i++) {
      fldsiz = new_field_sizes[i];
      k = 1;
      for (fldlen = 1; fldlen < 17; fldlen++) {
	k *= 2;
	if (k >= fldsiz)
	  break;
      }
      new_field_ends[i] = prev_end + fldlen;
      prev_end += fldlen;
    }
    for (i=0;i<numfields;i++) {
      field_sizes[i] = new_field_sizes[i];
      field_ends[i] = new_field_ends[i];
    }
    /*  3.  remap all indiv's in oldpop into new rep in newpop */
    for (j=0;j<popsize;j++) {
      for (i=0;i<numfields;i++) {
	start = old_map_to_std[i] > 0 ?
		  old_field_ends[old_map_to_std[i] - 1] + 1 : 1;
	end = old_field_ends[old_map_to_std[i]];
	tmp = ithruj2int(start, end, oldpop[j].chrom);
/*
	fprintf(outfp, "\n val. in new field[%d] = %d"
		       ,inverse_new_map_to_std[old_map_to_std[i]], tmp);
*/
	putfield(inverse_new_map_to_std[old_map_to_std[i]], numfields, tmp
		 , tempchrom);
      }
    for (i=0;i<chromsize;i++)
      oldpop[j].chrom[i] = tempchrom[i];
    }
 
    /*  4. free up the memory temporarily malloc'd here. */
    free (new_field_ends);
    new_field_ends = NULL;
    free (new_field_sizes);
    new_field_sizes = NULL;
    free (old_map_to_std);
    old_map_to_std = NULL;
    free (inverse_new_map_to_std);
    inverse_new_map_to_std = NULL;
    free (old_field_ends);
    old_field_ends = NULL;
    free (tempchrom);
    tempchrom = NULL;
 
    /*  This subpop now works in the "std" way, but with a new local map */
  }
 
  if (quiet < 3) {
	fprintf(outfp, "\n** New Inversion Map for Subpopulation %d:\n", popno);
	for (i=0;i<((numfields-1)/16) + 1;i++) {
	    for (j=0;j < 16 && i * 16 + j < numfields;j++) {
		fprintf(outfp, "%d ", map[16*i+j]);
	    }
	    fprintf(outfp, "\n");
	}
  }
}

