/*----------------------------------------------------------------------------*/
/* master.c -- Routines to read the .mst files, containing the definition of  */
/*             each subpopulation's neighboring subpopulations and the        */
/*             number and type of individuals (random or best) to be read     */
/*             from those files.  Also contains routines to actually do the   */
/*             reading of the "migrating" individuals from the neighbor files */
/*             at the beginning of each cycle (after first), and to read the  */
/*             chromosome's map_to_std of a donor population for use in       */
/*             inversion.                                                     */
/*----------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "external.h"

#define SEEK_SET         0
#define SEEK_CUR         1
#define MEMPTY           0
#define MMAX_NEIGHBOR   50  /* Increase if npops > 50 */
#define RETRIES 10

/* NOTE:  Next 3 defines are based on layout of the .ckp file.  If that */
/* file format is changed (in checkhdr.c), then THESE DEFINES MUST BE   */
/* modified accordingly.  Note: code where used below also adds         */
/* GetUtilitySize, ChromSize * sizeof(unsigned), and numfields * sizeof(int) */
/* to get the correct offset.                                                */
#define ChromSizePosn   2*sizeof(int)
#define ChromMapPosn    5*sizeof(int) 
#define SkipToFieldEnds 2*sizeof(int) 

static int NeighborTable[MMAX_NEIGHBOR][MMAX_NEIGHBOR][3];
static int SubpopCnt;
static int n_wanted_ind = 0;
static int incest_redn_ind = 1;
static int crowd_amt_ind = 2;
static int buf_pos;
static unsigned long mask, indexnum;

BOOL
ReadNeighborTable(filename)
char   *filename;
{
  FILE   *fp;
  int     i, j, k;
  int     receiving_pop, neighborcnt;
  int     neighbor, indwanted, migration_incest_reduction;
  int     migration_crowding_amount;
  BOOL    ShortForm, want_best;
  
  if (npops > MMAX_NEIGHBOR) {
    fprintf(stderr, "--** Previously entered value for number of populations,");
    fprintf(stderr, " %d, is too big for neighbor table.\n", npops);
    fprintf(stderr, "--** Increase MMAX_NEIGHBOR in file master.c and");
    fprintf(stderr, " recompile.\n--** Program exiting.\n\n");
    exit(-99);
  }
  
  for (i = 0; i < npops; i++)
    for (j = 0; j < npops; j++)
      for (k = 0; k < 3; k++)
	NeighborTable[i][j][k] = MEMPTY;
  /* Revising, dropping t from rt */ 
  if ((fp = fopen(filename, "r")) == NULL) {
    fprintf(stderr, "Error - Unable to open master file %s.\n", filename);
    return FALSE;
  }
  
  if (ffscanf("subcnt", fp, "%d", &SubpopCnt) != SUCC) {
    fprintf(stderr, "\nError in reading subcnt in file %s.", filename);
    fclose(fp);
    return FALSE;
  }
  
  ShortForm = FALSE;
  if (SubpopCnt < 0) {
    /* Using short form of master file specification */
    ShortForm = TRUE;
    SubpopCnt = -SubpopCnt;
  }
  
  if (SubpopCnt != npops) {
    fprintf(stderr, "\nnpops in master file isn't compatible with the run.");
    fclose(fp);
    exit(1);
  }
  
  for (i = 0; i < SubpopCnt; i++) {
    if (ffscanf2("subpop", fp, "%d %d", &receiving_pop, &neighborcnt) != SUCC) {
      fclose(fp);
      return TRUE;
    }
    if (receiving_pop < 0 || receiving_pop > MMAX_NEIGHBOR ||
	neighborcnt < 0 || neighborcnt > MMAX_NEIGHBOR) {
      fprintf(stderr, "\nError:receiving_pop is %d and neighborcnt is %d.", 
	      receiving_pop, neighborcnt);
      fclose(fp);
      return FALSE;
    }
    for (j = 0; j < neighborcnt; j++) {
      if (ffscanf4("neighbor", fp, "%d %d %d %d", &neighbor, &indwanted
		   , &migration_incest_reduction
		   , &migration_crowding_amount)
	  != SUCC) {
	fprintf(stderr, "\nError in neighbor in ReadNeighborTable.");
	fclose(fp);
	return FALSE;
      }
      if (neighbor < 0 || neighbor > MMAX_NEIGHBOR) {
	fprintf(stderr, "\nError:neighbor is %d.", neighbor);
	fclose(fp);
	return FALSE;
      }
      NeighborTable[receiving_pop][neighbor][n_wanted_ind]
	= indwanted;
      /* Note:  next two must stay zero if different_reps is TRUE. */
      if (!different_reps) {
	NeighborTable[receiving_pop][neighbor][incest_redn_ind]
	  = migration_incest_reduction;
	NeighborTable[receiving_pop][neighbor][crowd_amt_ind]
	  = migration_crowding_amount;
      }
    }
    if (ShortForm)
      break;
  }
  fclose(fp);
  /*  for(i=0;i<SubpopCnt;i++) {
      for(j=0;j<SubpopCnt;j++)
      for (k=0;k<3;k++) {
	fprintf(stderr, "%d ", NeighborTable[i][j][k]);
      }
      fprintf(stderr, "\n");
      }  */
  if (ShortForm) {
    for (i = 1; i < SubpopCnt; i++) {
      for (j = 0; j < SubpopCnt; j++)
	for (k = 0; k < 3; k++)
	  NeighborTable[i] [(j+1)%SubpopCnt][k]
	    = NeighborTable[i-1][j][k];
    }
  }
  /*  for(i=0;i<SubpopCnt;i++) {
      for(j=0;j<SubpopCnt;j++)
      for(k=0;k<3;k++)
      fprintf(stderr, "%d ", NeighborTable[i][j][k]);
      fprintf(stderr, "\n");
      }  */
  return TRUE;
}

void
ReadFromNeighbor(popno)
     /* Using the master file and subpopulation checkpoint files, read into */
     /* subpopulation popno any immigrants it is to receive from neighbors. */
     int     popno;
{
  char    ind_file_buf[14];
  int     i;
  BOOL    want_best;

  if (popno < 0 || SubpopCnt <= popno) {
    fprintf(stderr, "\nError: popno is %d and SubpopCnt is %d.", 
	    popno, SubpopCnt);
    return;
  }

  for (i = 0; i < SubpopCnt; i++) {
    if (NeighborTable[popno][i][n_wanted_ind] == MEMPTY)
      continue;

    CombineFileName(ind_file_buf, restartfileprefix, i, ".ind");
    /* buf is name of .ind file for neighbor to be read from. */
    /* If table has -m, read best and m-1 random guys; if m, just */
    /* read m random guys, if empty (==0), no immigrants from subpop i. */
    if (NeighborTable[popno][i][n_wanted_ind] < 0)
      want_best = TRUE;
    else
      want_best = FALSE;

    /* If using inversion, must first get the donor pop's inversion */
    /* map into global array donor_pop_map_to_std, for use in */
    /* function GetMigrants below.  Since fields may be different sizes, */
    /* must also get field_ends for donor pop here. */
    if (using_inversion) {
      if (!get_donor_pop_inversion_map_and_field_ends(i)) {
	fprintf(stderr, "\n***Failed to get donor pop inversion map");
	fprintf(outfp, "\n***Failed to get donor pop inversion map");
	exit(-1);
      }
    }
    
    if (GetMigrants(oldpop, i, ind_file_buf, 
		    abs(NeighborTable[popno][i][n_wanted_ind]), want_best,
		    NeighborTable[popno][i][incest_redn_ind],
		    NeighborTable[popno][i][crowd_amt_ind]) == FALSE) {
	fprintf(stderr, "\nError: Unable to get migrant from %s.",ind_file_buf);
	fprintf(stderr, "\nContinuing...\n");
      }
  }
}


/* Fetches the inversion map and field_ends from donor pop's .ckp file */
/* NOTE:  This code assumes that the size of the utility fields recorded */
/* to the checkpoint files is THE SAME for all subpopulations.  If this  */
/* is NOT true, you must modify the code below (or perhaps GetUtilitySize */
/* in utility.c) to capture the correct size for the donor subpopulation. */
int donor_pop_num;
BOOL
get_donor_pop_inversion_map_and_field_ends(donor_pop_num)
{
  /* ChromSize below is for donor pop, but should match this pop's for */
  /* inversion to work. */
  int  ChromSize, NumFields, i;
  FILE *fp;
  BOOL cmplt;
  char ckp_fname[13];
  long startofbuffer;

  CombineFileName(ckp_fname, restartfileprefix, donor_pop_num, ".ckp");

  if (!(fp = start_read(ckp_fname, 0, RETRIES, &mask, &indexnum, &buf_pos)))
    {
      fprintf(stderr, "\n*** ERROR: Unable to open %s for read ***\n", 
	      ckp_fname);
      fprintf(stderr, "\nRun will be terminated.\n");
      return (int)-1;           /* Flag for failure */
    }

  /* First skip through to the chromosome size, needed for further skip */
  startofbuffer = ftell(fp);
  if (fseek(fp, (long) ChromSizePosn, SEEK_CUR))
    return FALSE;
  if (fread(&ChromSize, sizeof(int), 1, fp) !=1)
    return FALSE;
  /* Next field is numfields (call local NumFields), for donor population, */
  /* so read it. */
  if (fread(&NumFields, sizeof(int), 1, fp) !=1)
    return FALSE;
  if (fseek(fp, (long)(ChromMapPosn + startofbuffer), SEEK_SET))
    return FALSE;
  /* The read below sets the global variable donor_pop_map_to_std[].  */
  if (fread(donor_pop_map_to_std, sizeof(int), NumFields, fp)!= NumFields)
    return FALSE;
  if (fseek(fp, (long) SkipToFieldEnds, SEEK_CUR))
    return FALSE;
  if (fread(donor_pop_field_ends, sizeof(int), NumFields, fp)!= NumFields)
    return FALSE;


  if ((end_read(ckp_fname, fp, 0, RETRIES, mask, indexnum, buf_pos, &cmplt))) {
    fp = NULL;
    return TRUE;
  }
  
  return FALSE;
}
