/*----------------------------------------------------------------------------*/
/* multimut.c -- Multi-bit spatially correlated mutation operator for ordinary*/
/*               (non-permutation) problems.  With probability p_multimut     */
/*               it randomly sets an adjacent set of 1 or more loci (bits or  */
/*               fields), where the number of loci randomly set is Poisson    */
/*               distributed with probability                                 */
/*                      1./avg_mut_width -- same as saying expected           */
/*               number of adjacent mutations is avg_mut_width.               */
/*               Treats chromosome as circular.  Written 6/96 by EDG.         */
/*               Note:  crude, slow -- could be speeded up with some          */
/*               attention, as I did ordinary mutation earlier.               */
/*----------------------------------------------------------------------------*/

#include "external.h"
#include <math.h>

char which_mutation[] = "multimut.c";

void
mutation(child)
/* This function called for BINARY CHROMOSOME CASE:  alpha_size = 2 */
unsigned *child;

{
    int start, end, change_length, mut_locus, wordnumber, bitinword, i;
    float prob_wider;
    unsigned int mask;

    /* TEMPORARILY, until revise code so load ALL operators and choose from */
    /* among all at run time, just use pmutation to run this one, instead.  */
    p_multimut = pmutation;

    /* First, decide whether or not to do a mutation for this chromosome,    */
    /* using p_multimut.                                                     */ 
    if (flip(1. - p_multimut)) 
	return;
    else {
	/* Doing one with exp width avg_mut_width, so figure out how many    */
	/* bits to change (may be 0) */
	prob_wider = (1. - 1./avg_mut_width);
	change_length = 0;
        while (flip(prob_wider)) {
	  change_length++;   
	}
	/* Note:  expectation breaks down if avg_mut_width nears numfields */
	if (change_length >= numfields) change_length = numfields;
	if (change_length == 0) return;
	/* Figure out where to start, stop the multi-bit mutation (at this */
	/* point, it is not 0-length).                                     */
        /* Pick start point (NOTE:  will wrap chrom around end): */
	start = rnd(0, numfields-1);
        end = start + change_length - 1;
	/* Do the mutation: */
	/* writechrom(child); */
        for (i=start; i<=end;i++) {
	  if (flip(0.5)) {
	    /* NOTE:  Do the bit flip only with 50% probability; otherwise, */
	    /* only get start substring or its complement -- too drastic. */
            mut_locus = i % numfields;
	    wordnumber = mut_locus/UINTSIZE;
	    bitinword = mut_locus%UINTSIZE;
	    mask = 1 << bitinword;
	    child[wordnumber] ^= mask;
	  }
	}
	nmutation++;
	/* writechrom(child); */
	/* fprintf(outfp, "\n"); */
    }
}

void
field_mutation(child)
unsigned *child;

/* Same idea as routine above, but for non-binary fields -- must keep each   */
/* of one or more fields within its allowable range, field_sizes[i].         */
/* Mutate chosen fields to uniform randomly chosen members of those fields'  */
/* legal alphabets, with probability p_multimut; count # of mutations.       */
/* NOTE:  probability of mutation is same for each field, regardless of its  */
/* length (not realistic in nature, but usual in GA).                        */
{
    int start, end, change_length, new_allele, mut_field, allele, i;
    float prob_wider;
    unsigned int mask;

    /* TEMPORARILY, until revise code so load ALL operators and choose from */
    /* among all at run time, just use pmutation to run this one, instead.  */
    p_multimut = pmutation;

    /* First, decide whether or not to do a mutation for this chromosome,    */
    /* using p_multimut.                                                     */ 
    if (!flip(p_multimut)) 
	return;
    else {
	/* Doing one with exp width avg_mut_width, so figure out how many    */
	/* fields to change (may be 0) */
	prob_wider = (1. - 1./avg_mut_width);
	change_length = 0;
        while (flip(prob_wider)) {
	  change_length++;   
	}
	/* Note:  expectation breaks down if avg_mut_width nears numfields */
	if (change_length >= numfields) change_length = numfields;
	if (change_length == 0) return;
	/* Figure out where to start, stop the multi-field mutation (at this */
	/* point, it is not 0-length).                                     */
        /* Pick start point (NOTE:  will wrap chrom around end): */
	start = rnd(0, numfields-1);
        end = start + change_length - 1;
	/* writefields(child); */
	/* Do the mutation: */
        for (i=start; i<=end;i++) {
          mut_field = i % numfields;
	  /* getfield numbers fields in {0, numfields-1} */
	  allele = getfield(mut_field, numfields, child);
	  /* Don't FORCE each field to change, just ALLOW it to. */
          new_allele = rnd(0, field_sizes[mut_field] - 1);
	  putfield(mut_field, numfields, new_allele, child);
	}
	nmutation++;
	/* writefields(child); */
	fprintf(outfp, "\n");
    }
}
