/*----------------------------------------------------------------------------*/
/* rnkslect.c - Rank-based selection                                          */
/*                                                                            */
/*                                                                            */
/*     Once the population is ranked, there are various ways to use this      */
/* information.  Here, we use linear ranking (Baker 1975) which gives this    */
/* formula for probability p of (ranked from 1..Lambda) individual i:         */
/*           1                                   i - 1                        */
/*   p(i) = --- * (nu_max - (nu_max - nu_min) * ------- )                     */
/*          Lam                                 Lam - 1                       */
/*                                                                            */
/*   where 1 <= nu_max <= 2, and nu_min = 2 - nu_max.                         */
/*                                                                            */
/*     Using the formula for the sum of the numbers from 1                    */
/* to N = (N + 1) * N / 2, this quickly reduces to 1 by simple algebra.       */
/* This formula gives a weighted value from nu_max = 2, wherein the term      */
/* of nu_max - nu_min reduces to 2, and thus the best individual has maximal  */
/* weight 2 times the mean weight, to nu_max = 1, (so nu_max - nu_min also    */
/* equals 1), so the individuals all have equal probability 1/Lam.            */
/*                                                                            */
/*     The actual selection used is suselect.c.  However, the selection is    */
/* summed over the rank-associated probabilities, not the fitnesses.          */
/*                                                                            */
/*     For most GA problems, the evaluation function uses 90+% of the         */
/* computational power, so the time spent (quick)sorting here is really       */
/* a minor consideration.                                                     */
/*----------------------------------------------------------------------------*/
#include "external.h"   /* Includes stdlib.h, where malloc() is. */

#define RANKDEFAULT 2.0 /* Must be between 1.0 and 2.0, inclusive.            */
int *ranklist = NULL;   /* Warning:  Make sure the type this array points to  */
                        /*           matches what you tell qsort below (not   */
                        /*           to mention malloc().)  Also, note that   */
                        /*           a compiler with 2-byte ints might have a */
                        /*           problem with pops of more than 32768     */
                        /*           individuals.                             */
float *ranklist_probabilities = NULL; /* Precalculated probabilities for ranked items. */                        

int *choices = NULL, nremain = 0;

char which_selection[] = "rnkslect.c";
                        
void
select_memory()  /* Malloc a block of mem for popsize ints = index into population. */
{                /* Also get a rank factor. */
    int i;
    char ch;
    float oneoverpopsize;
    size_t nbytes = 0;
    float rankfactor = RANKDEFAULT;

    if(scalemult > 1.) { /* This factor naturally serves this function when */
                        /* rank-based scaling is done, and it makes no sense */
                        /* to do window scaling or linear scaling, so it is */
                        /* used in this way. */
        rankfactor = scalemult;
    }
    else {
        fprintf(stderr, "\n***ERROR:  Window scaling and rank-based selection");
        fprintf(stderr, " are incompatible.\nParameter scalemult must be >");
        fprintf(stderr, " 1.0 to use rank-based selection.\nTurn off window");
        fprintf(stderr, " scaling and set");
        fprintf(stderr, " 1.0 < scalemult <=2.0\nor use a different form of");
        fprintf(stderr, " selection\n\n");
        exit(-1);
    } 

    nbytes = popsize * sizeof(int);
    if ((choices = (int *) malloc(nbytes)) == NULL)
	    nomemory("choices");

    if (ranklist_probabilities)
    	{
    	 free(ranklist_probabilities);
    	 ranklist_probabilities = NULL;
    	}
    /* Values filled in later. */
    if ((ranklist_probabilities = (float *) malloc((size_t) (((size_t) popsize) * sizeof(float)))) == NULL)
    	nomemory("ranklist_probabilities");
    	
    

    if (ranklist)  /* Should NEVER be a non-NULL garbage pointer. */
    	{
    	 free(ranklist);
    	 ranklist = NULL;
    	}
    nbytes = ((size_t) popsize) * sizeof(int);
    if ((ranklist = (int *) malloc(nbytes)) == NULL)
		nomemory("ranklist");

    for (i=0;i<popsize;i++)     /* Fill index array with indices into oldpop. */
 	    ranklist[i] = i;    /* Since pop's fitnesses are in random order, */
 	                        /* the indices here are random as far as the  */
 	                        /* population is concerned.  Thus we don't    */
 	                        /* have to worry about a possible qsort       */
 	                        /* worst case scenario (from this             */
 	                        /* initialization, anyway.)                   */


/*           1                                   i - 1                        */
/*   p(i) = --- * (nu_max - (nu_max - nu_min) * ------- )                     */
/*          Lam                                 Lam - 1                       */
/*                                                                            */
/*   where 1 <= nu_max <= 2, and nu_min = 2 - nu_max.                         */
    	
    oneoverpopsize = (float) (1.0 / (float) popsize);	
    if (ranklist_probabilities)	
    	for (i = 0;i<popsize;i++)
    		{
    		 ranklist_probabilities[i] =  /* Note: i is already "i - 1", and Lambda is popsize. */
    		      oneoverpopsize * (rankfactor - (rankfactor - (2.0 - rankfactor)) * ((float) i) / (((float) popsize) - 1));
    		}

}

void
select_free()
/* Here, you must free whatever you malloc above, done at end of each
subpopulation (Manypops) or run (Onepop). */
{
	if (ranklist != NULL)
    	free(ranklist); /* ANSI says its OK to free NULL in any event. */
    ranklist = NULL;    /* Clean up the pointer (now it's garbage.)     */
    if (ranklist_probabilities != NULL)
    	free(ranklist_probabilities);
    ranklist_probabilities = NULL;
    if (choices != NULL)
    	free(choices);
    choices = NULL;
}


/* Function: rank_compare_greater
   Description: Comparison function for qsort().
	   Qsort requires a function pointer argument.  It is a
	   callback function to us, the user, to tell qsort
	   which element is greater.  We return an
	   int whose value is < 0 if e1 is greater, 0 if
	   they are equal, and > 0 if e2 is greater.  It is
	   up to us to define what greater means.

   Input, Input Modifications, Return Values: Predefined by
       qsort() definition.  (see below) */
#ifdef PROTOTYPES_ACCEPTED
int
rank_compare_greater(const void *elem1, const void *elem2)
#else
int
rank_compare_greater(elem1, elem2)
void *elem1, *elem2;
#endif
{
  if (oldpop[*((int *) elem1)].fitness > oldpop[*((int *) elem2)].fitness)
    return(-1);
  else if (oldpop[*((int *) elem1)].fitness < oldpop[*((int *) elem2)].fitness)
    return(1);

  return(0);
}


void
preselect()
{   /* Any setup stuff goes here, done ONCE per generation.  Uses stochastic */
    /* universal sampling (like suselect.c), but based on linear ranking     */
    /* values rather than fitness values.                                    */

    int    i;
    int    j, k;
    float  pointer, sum;

    /* Faster to test once with two routines than to  */
    /* test within the rank_compare function itself.  */
    /* Remember: This is RANK based selection -- */
    /* fitness PER SE is never used!  Selection based */
    /* on relative POSITION only! */
    qsort(ranklist, (size_t)popsize, sizeof(int), rank_compare_greater);
    
    if (fit_avg == 0.) {
	    for (j = 0; j < popsize; j++)
	        choices[j] = j;
    }
    else {
      k = 0;
      pointer = randomperc();
      sum = 0.0;
      for (j = 0; j < popsize; j++)
	for (sum += ranklist_probabilities[j] * (float) popsize; 
	     sum > pointer; pointer++)
	  choices[k++] = j;
    }
    nremain = popsize - 1;

}



int
selection()
/* This is called ONCE PER INDIVIDUAL TO BE SELECTED */
{
	/*********************************************************/
	/*     The suselect.c method is used, but running on the */
        /* probabilities generated from the RANKING scheme, not  */
        /* on relative fitnesses, as happens when suselect.c is  */
        /* used alone as the selection method.                   */


    /* suselect.c method                                         */

    int     jpick, slect;
    jpick = rnd(0, nremain);
    slect = choices[jpick];
    choices[jpick] = choices[nremain];
    nremain--;
    return (slect);

}

