/*----------------------------------------------------------------------------*/
/* statistic.c - compute the fitness statistics                               */
/*----------------------------------------------------------------------------*/
/* Modified 6/96 (Release 3.2) to include Boltzmann scaling, EDG.             */

#include <math.h>
#include <limits.h>
#include "external.h"

static int eval_spec_converg;
static int tell_globalstats_to_report_this_guy = 0;

int local_bestfit_popno, local_bestfit_chromsize,local_bestfit_numfields;

void
statistics(pop)
/* Calculate population statistics, perform scaling based on those if desired,*/
#ifndef DBL_MAX
#define DBL_MAX 1.E308
#endif
/* and update some variables for individuals NOT evaluated at generation time.*/
/* NOTE:  keeping raw fitness in .init_fitness, then calculate scaled fitness,*/
/* and keep it in .fitness.  */

struct individual *pop;
{
    int     record_new_best_fitness = 0;
    /* "local", as used below, means among all subpops run by THIS process. */
    int     record_new_local_best_fitness = 0;
    int     i, j, best;
    double  sumsquares, variance;

    sumfitness = 0.0;
    sumsquares = 0.0;
    minraw = pop[0].init_fitness;
    if (minraw < 0.)
	minraw = 0.;
    fit_max = pop[0].init_fitness;

    /* Loop for fit_max, minraw, sumfitness */
    bestnow = 0;
    for (j = 0; j < popsize; j++) {
	if (pop[j].init_fitness < 0.)
	    pop[j].init_fitness = 0.;
	sumfitness = sumfitness + pop[j].init_fitness;  /* Accumulate */
	sumsquares = sumsquares + pop[j].init_fitness
				* pop[j].init_fitness;

	if (pop[j].init_fitness > fit_max) {
	    fit_max = pop[j].init_fitness;      /* New fit_max */
	    bestnow = j;
	}
	if (pop[j].init_fitness < minraw)
	    minraw = pop[j].init_fitness;
	/* New min */

	/* new best-fit individual of all time (this population) */
	if (pop[j].init_fitness > bestfit.init_fitness) {
	    best = j;   /* index of new best individual ever */
	    /* flag saying record best (scaled) fitness below */
	    record_new_best_fitness = 1;   
	    for (i = 0; i < chromsize; i++)
		bestfit.chrom[i] = pop[best].chrom[i];

	    bestfit.init_fitness = pop[best].init_fitness;
	    bestfit.neval = pop[best].neval;
	    bestfit.generation = gen;
	    for (i = 0; i < numfields; i++) {
		bestfit.map_to_std[i] = this_pop_map_to_std[i];
		bestfit.field_ends[i] = field_ends[i];
	    }
	    sys_util_copy_utility(bestfit.utility, pop[best].utility);
	}
	/* if new best-fit individual is a new best among all local pops -- */
	/* (for THIS process(or)), record him as a new local_bestfit.    */
	if (bestfit.init_fitness > local_bestfit.init_fitness) {
	    local_bestfit.init_fitness = bestfit.init_fitness;
	    if (pop == newpop)  /* uncomment if don't want initial guys */
				/* to be reported as new bests. */
		tell_globalstats_to_report_this_guy = 1;
	    record_new_local_best_fitness = 1;
	    for (i = 0; i < chromsize; i++)
		local_bestfit.chrom[i] = bestfit.chrom[i];
	    local_bestfit.generation = bestfit.generation;
	    /* Record neval (among all pops, all processors) for new best */
	    /* among all LOCAL pops (this process), correcting to proper  */
	    /* value for when it actually occurred. */
	    local_bestfit.neval = all_pops_neval + local_cycle_neval;
	    for (i = 0; i < numfields; i++) {
		local_bestfit.map_to_std[i] = this_pop_map_to_std[i];
		local_bestfit.field_ends[i] = field_ends[i];
	    }
	    sys_util_copy_utility(local_bestfit.utility, bestfit.utility);
	    /* Record best-fit's popno, chromsize, numfields */
	    local_bestfit_popno = popno;
	    local_bestfit_chromsize = chromsize;
	    local_bestfit_numfields = numfields;
	}
    }

    /* Calculate average unscaled fitness */
    avgraw = sumfitness / popsize;

    /* Track data needed to calculate on-line performance, off-line */
    /* performance for this population and local populations to this   */
    /* processor (ultimately used to track all populations on all */
    /* processors. */

    if(pop == newpop || gen == 0) {
	/* Increment counters, statistics if NOT initializing or if gen=0 */
	/* This avoids using the same population twice (at end of one cycle */
	/* and beginning of the next) in these measures.  (The stored     */
	/* measures can still be printed when restarts are done.)         */

	one_pop_cum_sumfitness += sumfitness;
	one_pop_sum_best_fitness += pop[bestnow].init_fitness;
	one_pop_best_fitness_count += 1;
	one_pop_online_denominator += popsize;
	local_cycle_sumfitness += sumfitness;
	local_cycle_sum_best_fitness += pop[bestnow].init_fitness;
	local_cycle_best_fitness_count +=1;
	local_cycle_online_denominator += popsize;
    }
	/* Now, for printing for THIS pop only, */
	on_line_performance = one_pop_cum_sumfitness
			      / one_pop_online_denominator;
	off_line_performance = one_pop_sum_best_fitness
			      / one_pop_best_fitness_count;

	/* Now, also update the LOCAL version of all_pops measures, */
	/* BUT REMEMBER, that these won't ever be written for other pops */
	/* to use... Instead, the local_cycle stats will be combined with */
	/* the LATEST all_pops stats FRESHLY READ from the .stt file */
	/* JUST BEFORE this processor adds this cycle's contributions to */
	/* that file.  Therefore, these variables set below are really   */
	/* just TEMPORARY, working versions of the variables which will  */
	/* be accumulated across all populations, but are useful for  */
	/* printing in globalstats(), for example.                 */

	all_pops_sumfitness += sumfitness;
	all_pops_sum_best_fitness += pop[bestnow].init_fitness;
	all_pops_best_fitness_count += 1;
	all_pops_online_denominator += popsize;
	all_pops_on_line_performance = all_pops_sumfitness
				      / all_pops_online_denominator;
	all_pops_off_line_performance = all_pops_sum_best_fitness
				      / all_pops_best_fitness_count;

    /* record for use if window scaling is turned on (even if off now) */
    savedmins[windowend % MAXSCALINGWINDOW] = minraw;

    /* If using special convergence measure or sigma truncation, or the */
    /* autocontrol feature for selection with Boltzmann scaling, must   */
    /* calculate std. deviation */

    eval_spec_converg = (conv_sigma_coeff != 7. && convinterval != 0);
    eval_spec_converg = (eval_spec_converg && gen % convinterval == 0);
    if (autocontrol_selection || sigma_trunc != 0.0 || eval_spec_converg) {
       variance = fabs(sumsquares / popsize - avgraw * avgraw);
       /* absolute value is just to assure that -0.0000000x does not cause a
       /* problem... can occur when population is completely converged */
	sigma = sqrt(variance);
    }
    if (autocontrol_selection) {
	/* Here is where the adjustment of beta is made to maintain */
	/* constant selection pressure, regardless of std. dev. sigma */
	/* of current distribution of fitnesses (see Shapiro & */
	/* Pruegel-Bennett for explanation) */ 
	beta = beta_s_tilde / sigma;
    }
					       
    /* Decide if need to calc convergence measures, do if must */
    if (convinterval != 0)
	convergemeasure(pop);   /* calculate convergence measures */

    if (beta >= 0. || autocontrol_selection) 
	Boltzmann_scale(pop);
    else if (scaling_window > -1)
	window_scaling(pop);
    else if (sigma_trunc != 0.0)
	sigma_truncation(pop);

    else { /* didn't do Boltzmann or window scaling or sigma truncation, so */
    /* just prepare for poss lin scaling or no scaling at all.  */
	for (j = 0; j < popsize; j++)
	    pop[j].fitness = pop[j].init_fitness;
	fit_min = minraw;
	fit_avg = avgraw;
    }

    /* Determine whether or not to use linear scaling; if not, leave */
    /* fitnesses at the raw or sigma-truncated values.   */
    if (scalemult > 1.0 && !autocontrol_selection 
	   && beta < 0. && !window_scaling)
	linear_scale(pop);

    /* Record scaled fitness of best ever individual if changed */
    if (record_new_best_fitness == 1)
	bestfit.fitness = pop[best].fitness;
    if (record_new_local_best_fitness == 1)
	local_bestfit.fitness = bestfit.fitness;
    /* get application dependent stats */
    app_stats(pop);
}


void
convergemeasure(pop)
/* calculates various measures of population convergence. */

struct individual *pop;
{
    int     end, bestnotclosest, j, k, distfrombest;
    float   fitness_cutoff;
    int     n_lost, n_converged;
    unsigned long bias;

    end = 0;
    bestnotclosest = 0;
    fitness_cutoff = avgraw + conv_sigma_coeff * sigma;

    if(eval_spec_converg) {
    /* First, calculate "special" convergence measure, diagnosing presence */
    /* of clones or near-copies of the best individual among high-fitness */
    /* "good" individuals, with user-defined std. dev. threshold for "good" */
    /* This measure is calculated and printed only at the user-specified */
    /* "convergence" interval */
	for (j = 0; j < popsize; j++)
	    if (pop[j].init_fitness >= fitness_cutoff)
		available[end++] = j;  /* reuse global list "available" */
	/* for indices of guys above cutoff */
	for (j = 0; j < end; j++) {
	    distfrombest = hamming(&(pop[available[j]]), &(pop[bestnow]));
	    for (k = 0; k < end; k++)
		if (hamming(&(pop[available[k]]), &(pop[available[j]]))
			< distfrombest) {
		    bestnotclosest++;
		    break;
		}
	}
	if (end != 0)
	    conv_measure = 100. * (1. - (float)bestnotclosest / end);
	else
	    conv_measure = -99.;
    }
    /* Now calculate the traditional convergence measures */
    n_lost = 0;
    n_converged = 0;
    bias = 0;
    for (k = 0; k < lchrom; k++)
	ones[k] = 0;

    for (j = 0; j < popsize; j++)
	count_ones_at_loci(pop[j].chrom);

    for (k=0;k<lchrom;k++) {
	if(ones[k] == popsize || ones[k] == 0)
	    n_lost += 1;
	if(ones[k] >= ceil(0.95 * popsize) || ones[k] <= (int)0.05 * popsize) {
	    n_converged += 1;
	}
	bias += (ones[k]>(popsize/2)) ? ones[k] : (popsize - ones[k]);
    }
    percent_lost = (100. / lchrom) * n_lost;
    percent_converged = (100. / lchrom) * n_converged;
    avg_bias = 100. * ((float) bias / ((float)popsize * lchrom));
}



void
count_ones_at_loci(chrom)
/* Calculate the number of ones in each locus of a chromosome   */
/* Note that the most significant bit of the chromosome is the  */
/* RIGHTMOST bit, not the leftmost bit, as would be expected... */
/* Added by EDG, October, 1993 */

unsigned *chrom;
{
    int     j, k, stop, pos;
    unsigned mask = 1, tmp;

    for (k = 0; k < chromsize; k++) {
	tmp = chrom[k];
	if (k == (chromsize - 1))
	    stop = lchrom - (k * UINTSIZE);
	else
	    stop = UINTSIZE;

	for (j = 0; j < stop; j++) {
	    if (tmp & mask) {
		pos = k * UINTSIZE + j;
		ones[pos] += 1;
	    }
	    tmp = tmp >> 1;

	}
    }
}


void
sigma_truncation(pop)
/* Perform sigma truncation, per Goldberg book,p.125. NOTE: does NOT preserve */
/* mean fitness, so do not use if that is necessary.                          */

struct individual *pop;
{
    double  sumtruncatedfitnesses = 0.;
    int     j;

    for (j = 0; j < popsize; j++) {
	pop[j].fitness = pop[j].init_fitness - (avgraw - sigma_trunc * sigma);
	if (pop[j].fitness < 0.0)
	    pop[j].fitness = 0.0;
	sumtruncatedfitnesses += pop[j].fitness;
    }
    fit_max = fit_max - (avgraw - sigma_trunc * sigma); /* fit_max can be
						 * freely scaled, since raw
						 * fit_ max value is available
						 * in init_fitness[bestnow] */
    if (fit_max < 0.)
	fit_max = 0.;
    fit_min = minraw - (avgraw - sigma_trunc * sigma);
    if (fit_min < 0.0)
	fit_min = 0.0;
    fit_avg = sumtruncatedfitnesses / popsize;  /* new fit_avg must be exact
						 * for some selection methods */
    if (fit_avg < 0.)
	fit_avg = 0.;
    sumfitness = sumtruncatedfitnesses;
}


void
Boltzmann_scale(pop)
/* Perform Boltzmann scaling, as described in Shapiro and Pruegel-Bennett */
/* papers.  NOTE: does NOT preserve mean fitness, so do not use */
/* if that is necessary.  Boltzmann scaling is especially useful when one */
/* needs to control selection pressure with a single parameter (beta), and */
/* is used in GALOPPS when "autocontrol_selection" is chosen (again, see */
/* the web site of Jonathan Shapiro for relevant papers).  This routine also */
/* sets beta from beta_s_tilde, IF autocontrol_selection is TRUE. */

/*   Recommended selection in GALOPPS is a Boltzmann scaling of fitnesses, */
/* followed by deterministic selection using Stochastic Uniform Sampling.  */
/* For discussion, see papers by Jonathan Shapiro, et al., or M. De La Maza */
/* and B. Tidor, "Increased flexibility in genetic algorithms:  The use of */
/*   variable Boltzman selective pressure to control propagation," in Proc.  */
/*   ORSA CSTS Conf:  Computer Science and Operations Research:  New         */
/*   Developments in their Interfaces, 1991.                                 */
/*                                                                           */
/*   This method  yields a selection mechanism     */
/*   with a SINGLE parameter which determines intensity of selection, over    */
/*   a very broad range.  Setting beta to 0 yields uniform sampling (i.e.,    */
/*   a very broad range.  Setting beta to 0 yields uniform sampling (i.e.,    */
/*   independent of fitness).  As beta approaches infinity, only the best     */
/*   individual is selected.  Setting beta small (in the vicinity of 0.01,    */
/*   for example) yields something akin to fitness-proportional sampling with */
/*   Stochastic Uniform Sampling.  Note that this selection is invariant      */
/*   under addition of a constant to all fitnesses. */
/*   If the user elects "automatic control of selection" by setting the       */
/*   parameter "autocontrol_selection" to TRUE in the input (or input         */
/*   file), then beta is controlled by the program as follows:        */
/*     1) beta_s_tilde is initialized to a user-specified value, which        */
/*        defaults to 0.5 * sqrt(2.0*log(popsize)) (for the subpopulation in  */
/*        which the selection is occurring), (data from Shapiro et al.)       */
/*     2) at each generation, beta is set to beta_s_tilde divided by the      */
/*        standard deviation of the current fitness distribution. Boltzmann   */
/*        scaling is done with this beta, and  this yields */
/*        nearly uniform "selection pressure" throughout the run.    */
/*     3) the scaled values are used in the SUS (Stochastic Uniform Sampling)*/
/*        selection method loaded by the user, suselect.c, the "deterministic"*/
/*        selection method preferred by Shapiro, et al. */

/*   If autocontrol_selection is FALSE, then the user-specified beta is       */
/*   simply used to calculate the selection probabilities for whatever */
/*   selection the user has chosen (roulette wheel, SUS, remainder selection, */
/*   stochastic remainder selection, etc.  (Of course, no scaling is useful in*/
/*   tournament selection.  */
/*   See Shapiro's web site for papers on how to control selection, which is  */
/*   well modeled in his statistical mechanics-based (or maximum entropy-     */
/*   based) approach, and behaves in a problem-independent manner.            *j

/*   Probabil. of sampling ith indiv. is proportional to:                     */
/*           (e**(beta * raw_fitness)) / Z,                                   */
/*   where Z = sum of those values for ALL i's.  (i.e., this is  */
/*   Boltzmann-weighted fitness).  Beta controls the intensity of selection.  */
/*   NOTE:  This is NOT the same as "Boltzmann Tournament Selection" as       */
/*         proposed by Goldberg, 1991.                                        */
/*   Coded for GALOPPS by Erik Goodman, January, 1996.                        */

struct individual *pop;

{
    int     j;
    double sum_beta_adj_fitnesses, tmp;
    double max_Boltzmann_expon;

    /* Calculate Boltzmann-scaled .fitnesses from raw .init_fitnesses. */
    max_Boltzmann_expon = log(DBL_MAX / (double)popsize);
    sum_beta_adj_fitnesses = 0.;                          
    for (j = 0; j < popsize; j++) {
	if ((tmp = (double) beta * (pop[j].init_fitness - avgraw)) 
	     > max_Boltzmann_expon) {
	  fprintf(stderr, "\n*** ERROR in Boltzmann scaling... cannot take ");
	  fprintf(stderr, "\n   exp() of scaled term, value is %f\n"
		  , j, tmp);
	  fprintf(stderr, "\n Either rescale fitness function, reset beta, ");
	  fprintf(stderr, "\n or do not use Boltzmann scaling.  Exiting.\n");
	  exit(-1);
	}
	/* fprintf(outfp, "\n tmp = %f", tmp); */
	pop[j].fitness = exp(tmp); /* using as a temporary until divide below */
	sum_beta_adj_fitnesses += pop[j].fitness;
    }
    for (j = 0; j < popsize; j++)  
	pop[j].fitness /= sum_beta_adj_fitnesses;
    fit_max = pop[bestnow].fitness;             /* fit_max can be
						 * freely scaled, since raw
						 * fit_ max value is available
						 * in init_fitness[bestnow] */
    if (fit_max < 0.)     /* safety only, should not happen */
	fit_max = 0.;
    fit_min = exp((double)beta * (minraw - avgraw)) 
		  / sum_beta_adj_fitnesses;
    if (fit_min < 0.0)
	fit_min = 0.0;
    fit_avg = 1.0 / popsize;  
    sumfitness = 1.0;
}


void
linear_scale(pop)
/* Linear scaling function... calculates (scaled) fitness values from         */
/* raw or sigma-truncated fitness values                                      */
/* It is an implementation of Goldberg's suggested linear scaling process from*/
/* Chapter 3, and preserves average fitness precisely.                        */

struct individual *pop;

{
    double  a, b, delta;
    int     j;

    /* Calc which scaling to use (to max or min fitness), to get coeffs */
    /* Test to avoid fitness<0 */
    if (fit_min > (scalemult * fit_avg - fit_max) / (scalemult - 1.0)) {
	/* scale so max fitness is scalemult * mean, mean maintained */
	delta = fit_max - fit_avg;
	if (delta >= 0.000001 * fit_avg) {
	    a = (scalemult - 1.0) * fit_avg / delta;
	    b = fit_avg * (fit_max - scalemult * fit_avg) / delta;

	}
	else {  /* completely converged, prevent 0 divide */
	    a = 1.0;
	    b = 0.0;
	}
    }
    else {
	/* scale so min fitness is 0.0, mean maintained */
	delta = fit_avg - fit_min;
	if (delta > 0.000001 * fit_avg) {
	    a = fit_avg / delta;
	    b = -fit_min * fit_avg / delta;
	}
	else {  /* completely converged, prevent 0 divide */
	    a = 1.0;
	    b = 0.0;
	}
    }

    /* Do actual scaling here, setting .fitness values to scaled ones */

    for (j = 0; j < popsize; j++) {
	pop[j].fitness = a * pop[j].fitness + b;
	if (pop[j].fitness < 0.)
	    pop[j].fitness = 0.;
    }
    fit_min = a * fit_min + b;
    if (fit_min < 0.)
	fit_min = 0.;
    fit_max = a * fit_max + b;  /* Note: fit_avg should be preserved, need
				 * not scale. */
    if (fit_max < 0.)
	fit_max = 0.;
}

void
window_scaling(pop)
/* Implements a scaling window which elevates floor of fitness scale (0) to */
/* lowest individual seen in scaling_window generations. */
struct individual *pop;
{
    int     j, k, togo;
    double  floor;

    windowend = windowend % MAXSCALINGWINDOW;
    togo = (windowend >= windowstart) ?
     (windowend - windowstart) : (windowend + MAXSCALINGWINDOW - windowstart);
    if (togo >= scaling_window + 1)
	windowstart = (windowstart + 1) % MAXSCALINGWINDOW;
    floor = savedmins[windowstart];

    for (j = windowstart; j != windowend; j = (j + 1) % MAXSCALINGWINDOW) {
	k = (j + 1) % MAXSCALINGWINDOW;
	floor = (savedmins[k] < floor) ? savedmins[k] : floor;  /* min */
    }
    /* do the actual window scaling */
    for (j = 0; j < popsize; j++) {
	pop[j].fitness = pop[j].init_fitness - floor;
	if (pop[j].fitness < 0.)
	    pop[j].fitness = 0.;
    }
    fit_avg = avgraw - floor;
    if (fit_avg < 0.)
	fit_avg = 0.;   /* handles completely converged case */
    fit_max = fit_max - floor;
    if (fit_max < 0.)
	fit_max = 0.;
    if (fit_avg > fit_max) /* possible due to truncation error */
	fit_avg = fit_max;
    fit_min = minraw - floor;
    if (fit_min < 0.)
	fit_min = 0.;
}

void
globalstats()
/* Reports any NEW "local" (THIS process or processor) best individual ever, */
/* and other performance statistics, among the local subpopulations on this  */
/* process or processor, for multi-pop runs, at end of each population's     */
/* execution during each cycle, if a new best of cycle individual was        */
/* found (and not in quiet mode).  However, NOTE that the neval and the      */
/* performance measures reported are RELATIVE TO ALL POPS on ALL PROCESSORS. */
{
    int     i, j, k;

    if (tell_globalstats_to_report_this_guy == 1) {
	tell_globalstats_to_report_this_guy = 0;
	if (quiet < 3) {
	    fprintf(outfp, "\n*** Subpop %d Achieved New Best Ever Among Subpops %d - %d:"
		    , popno, startpopnum, finishpopnum);
	    fprintf(outfp, "\n    Its Neval (rel. to all pops, all processors) = %ld *** \n"
		    , local_bestfit.neval);
	    fprintf(outfp, " raw fitness = %20.10f, gen = %d, chromosome =\n"
		    , local_bestfit.init_fitness, local_bestfit.generation);
	    writechrom(local_bestfit.chrom);
	    if (alpha_size != 2) writefields(local_bestfit.chrom);

	}

	app_new_global_best_report(); /* Do even if quiet == 3; user controls */

	if (quiet < 3) {

	    if (using_inversion) {
		fprintf(outfp, "* Inversion Map for That Individual:\n");
		for (i=0;i<((numfields-1)/16) + 1;i++) {
		    for (j=0;j < 16 && i * 16 + j < numfields;j++) {
			fprintf(outfp, "%d ", local_bestfit.map_to_std[16*i+j]);
		    }
		    fprintf(outfp, "\n");
		}
		fprintf(outfp
			, "That Chrom., UNMAPPED to 'standard' form is:\n");
		/* Use kidholder[0].chrom as a scratch chromosome. */
		map_chromosome_to_std_uneq_flds(local_bestfit.chrom, local_bestfit.map_to_std
				      , kidholder[0].chrom, local_bestfit.field_ends);
		writechrom(kidholder[0].chrom);
		if (alpha_size != 2)
		  writefields(kidholder[0].chrom);
	    }

	    fprintf(outfp,"\n\nAt END of subpopulation %d's run in cycle %d\n"
			 , popno, cycle);
	    fprintf(outfp
		   , "On-Line Perf. = %20.10f, Off-Line Perf. = %20.10f\n"
		   , all_pops_sumfitness / all_pops_online_denominator
		   , all_pops_sum_best_fitness / all_pops_best_fitness_count);
	}
    }
}


void
sys_util_stats(pop)
/* The call follows each generation, and is use to update any UTILITY */
/* values for the bestever structs, by copying them from the individuals  */
/* which still contain them, for recording and reporting.  */

struct individual *pop;
{
    static long savepreveval;  /* will keep eval num of prev global best indiv */
    int i, nbytes;
    unsigned char *destbyte, *sourcebyte;

    if (bestfit.neval != savepreveval) {
      /* Must record these if they changed this gen. Bestnow is valid as */
      /* best ever ONLY during the generation when it changed.           */
      nbytes = GetMaxUtilitySize();
      destbyte = (unsigned char *) bestfit.utility;
      sourcebyte = (unsigned char *) pop[bestnow].utility;
      for (i=0;i<nbytes; i++)
	destbyte[i] = sourcebyte[i];
      savepreveval = bestfit.neval;
    }
}

