/*************************************************************************
* 
*    user_in.c
* 
*    file of routines to accept user input, interactive or from files, using
*    command line overrides
* 
*    routines:
*    int commandline(int, char**)
*    void global_overrides()
*    void get_pop_overrides()
*    void GetUniversalInputs()
*    static void SetUpMigration()
*    static void RunWhichPops()
*    void subpop_data(BOOL*)
*    void problem_inputs()
*    int pop_inputs(BOOL*)
*
*    Modified so can specify an output file, but input from keyboard,EDG,4/9/96
*    Changed several defaults to legal (sensible) values, EDG, 4/9/96
*    Added input for Boltzmann selection w/wo automatic scaling, EDG, 4/9/96
*    
*    includes code modified from 
*    lil-gp Genetic Programming System, version 0.9b
*      Douglas Zongker       (zongker@cps.msu.edu)
*      William F. Punch III   (punch@cps.msu.edu)
* 
*    static void add_parameter(char*, char*, int)
*    static char *get_parameter(char*, char*)
*    static int parse_one_parameter(char*)
*    static int trim_string(char*)
*    static int delete_parameter(char*)
* 
*    5-95
*    Erik D. Goodman (goodman@egr.msu.edu)
*    Leslie Thomas Hoppensteadt  (hoppen@egr.msu.edu)
* 
*    (c) 1993, 1994, 1995, 1996 Board of Trustees        
*           Michigan State University           
*/
#include <stdio.h>
#include <time.h>
#include <math.h>
#include <string.h>
#include "external.h"

#define MAXPARAMLINELENGTH    255
#define PARAM_COPY_NONE   0
#define PARAM_COPY_NAME   1
#define PARAM_COPY_VALUE  2
#define PARAMETER_MINSIZE       31
#define PARAMETER_CHUNKSIZE     16

typedef struct _parameter
{
  char *n;
  char *v;
  int copyflags;
} parameter;

static parameter *param;
static int param_size = 0, param_alloc = 0;


/* read commandline and any switches */
int commandline(argc, argv)
int argc; 
char **argv;
{
  void usage();
  int i, changes;
  int errorflag = 0;
  char *inputname = 0;
  char *outputname = 0;
  char *extname = 0;
  char *transfer = 0;

  /* init */
  changes = 0;
  numfiles = argc -1;  /* obsolete, use interactive */
  interactive = 1;
  extending = 0;
  infp = stdin;
  outfp = stdout;
  restartfileprefix[0] = '\0';
  checkptfileprefix[0] = '\0';
  masterfileprefix[0] = '\0';

  plotfp = fopen("plot", "w");

  for (i = 1; i < argc; i++) {
    if (argv[i][0] != '-' || argv[i][2] != '\0') {
      fprintf(stderr, "Unrecognized command line option: \"%s\".\n", argv[i]);
      errorflag++;
      continue;
    }
    
    switch (argv[i][1]) {
    case 'i':
      if (++i < argc) {
	if (outputname && !(strcmp(outputname, argv[i]))) {
	  fprintf(stderr, "Input file can't be the same as output file.\n");
	  errorflag++;
	}
	else if (!(infp = fopen(argv[i], "r"))) {
	  fprintf(stderr, "Input file %s cannot be read.\n", argv[i]);
	  errorflag++;
	}
	else {
	  interactive = 0;
	  inputname = argv[i];
	}
      }
      else
	errorflag++;
      break;     
      
    case 'o':
      if (++i < argc) {
	if (inputname && !(strcmp(inputname, argv[i]))) {
	  fprintf(stderr, "Input file can't be the same as output file.\n");
	  errorflag++;
	}
	else if (!(outfp = fopen(argv[i], "w"))) {
	  fprintf(stderr, "Output file %s cannot be written.\n", argv[i]);
	  errorflag++;
	}
	else {
	  outputname = argv[i];
          fclose(plotfp);
          transfer = argv[i];
          plotfp = fopen(strcat(transfer, "plot"), "w");
        }
      }
      else
	errorflag++;
      break;     
      
    case 'c':
      if (++i < argc) {
	if (strlen(checkptfileprefix)) {
	  fprintf(stderr, "Can't use multiple checkpoint file prefixes.\n");
	  errorflag++;
	  continue;
	}

	strcpy(checkptfileprefix, argv[i]);
	checkptfileprefix[6] = '\0';
      }
      else
	errorflag++;
      break;
      
    case 'e':
      if (++i < argc) {
	if (extending) {
	  fprintf(stderr, "Can't start from multiple checkpoint files.\n");
	  errorflag++;
	  continue;
	}

	strcpy(restartfileprefix, argv[i]);
	restartfileprefix[6] = '\0';
	extending = 1;
      }
      else
	errorflag++;
      break;

    case 'm':
      if (++i < argc) {
	if (strlen(masterfileprefix)) {
	  fprintf(stderr, "Can't use multiple migration files.\n");
	  errorflag++;
	  continue;
	}

	strcpy(masterfileprefix, argv[i]);
	masterfileprefix[8] = '\0';
      }
      else
	errorflag++;
      break;

    case 'p':
      if (++i < argc) {
	changes++;
	if (!parse_one_parameter(argv[i])) {
	  errorflag++;
	  fprintf(stderr, "Malformed parameter: \"%s\".\n", argv[i]);
	}
      }
      else
	errorflag++;
      break;

    default:
      fprintf(stderr, "Unrecognized command line option: \"%s\".\n", argv[i]);
      errorflag++;
      break;
    }
  }
  
  if (infp)
    if (infp != stdin) 
      fprintf(outfp, "\nInput for this run read from %s.\n", inputname);
    else {
      fprintf(outfp, "\n***Running interactively, so command line \n");
      fprintf(outfp, "     input except for OUTPUT FILE NAME is not allowed.");
    }
    if (infp == stdin && changes) {
      fprintf(stderr, "You specified parameter overrides but no input file.\n");
      fprintf(stderr, "Command line parameter overrides are legal only when\n");
      fprintf(stderr, "using an input file.\n");
      errorflag++;
    }

  if (errorflag) {
    fprintf(stderr, "%d command line error(s) occurred.  Run not possible.\n\n",
	    errorflag);
    usage();
    exit(errorflag);
  }

  return changes;
}

void add_parameter(name, value, copyflags)
char *name, *value;
int copyflags;
{
  delete_parameter(name);
  
  while (param_alloc < param_size+1) {
    param_alloc += PARAMETER_CHUNKSIZE;

    /* lines below revised by EDG, 10/5/95, for compatibility with compilers */
    /* in which realloc() called with null pointer does not act like malloc() */
    if (param == NULL)
	param = (parameter *)malloc(param_alloc * sizeof(parameter));
    else
	param = (parameter *)realloc(param, param_alloc * sizeof(parameter));
  }
  
  if (copyflags & PARAM_COPY_NAME) {
    param[param_size].n = (char *)malloc(strlen(name)+1);
    strcpy(param[param_size].n, name);
  }
  else
    param[param_size].n = name;
  
  if (copyflags & PARAM_COPY_VALUE) {
    param[param_size].v = (char *)malloc(strlen(value)+1);
    strcpy(param[param_size].v, value);
  }
  else
    param[param_size].v = value;
  
  param[param_size].copyflags = copyflags;
  ++param_size;
}

char *get_parameter(name, n)
char *name, *n;
{
  int i;
  char buf[80];

  strcpy(buf, name);

  if (n) 
    strcat(buf, n);

  for (i = 0; i < param_size; ++i)
    if (!strcmp(buf, param[i].n))
      return param[i].v;
  
  return NULL;
}

int parse_one_parameter(buffer)
char *buffer;
{
  char name[MAXPARAMLINELENGTH+1];
  char data[MAXPARAMLINELENGTH+1];
  int i, j, k, l;
  int n, d;
  
  l = strlen(buffer);
  for (i = 0; i < l; ++i)
    if (buffer[i] == '#' || buffer[i] == ';'){
      buffer[i] = 0;
      break;
    }
  
  k = -1;
  j = 0;
  l = strlen(buffer);
  for (i = 0; i < l; ++i) {
    j += (buffer[i] != ' ' && buffer[i] != '\t' && buffer[i] != '\n');
    if (buffer[i] == '=') {
      k = i;
      strncpy(name, buffer, k);
      name[k] = 0;
      strcpy(data, buffer+k+1);
      break;
    }
  }
  
  if (k == -1)
    return 0;
  
  n = trim_string(name);
  d = trim_string(data);
  
  if (n == 0 || d == 0)
    return 0;
  else
    add_parameter(name, data, PARAM_COPY_NAME|PARAM_COPY_VALUE);
  
  return 1;
}

int trim_string(string)
char *string;
{
  int i, j, l;

  j = -1;
  l = strlen(string);
  
  for (i = 0; i < l; ++i)
    if (j == -1) {
      if (string[i] != ' ' && string[i] != '\t' && string[i] != '\n') {
	j = i;
	--i;
      }
    }
    else
      string[i-j] = string[i];
  
  if (j == -1) {
    string[0] = 0;
    return 0;
  }
  
  string[i-j] = 0;
  l = i-j;
  j = -1;
  
  for (i = 0; i < l; ++i)
    if (string[i] != ' ' && string[i] != '\t' && string[i] != '\n')
      j = i;
  
  string[j+1] = 0;
  return j+1;
}

int delete_parameter(name)
char *name;
{
  int i;
  
  for (i = 0; i < param_size; ++i)
    if (strcmp(name, param[i].n) == 0) {
      if (param[i].copyflags & PARAM_COPY_NAME)
	free(param[i].n);

      if (param[i].copyflags & PARAM_COPY_VALUE)
	free(param[i].v);

      if (param_size-1 != i) {
	param[i].n = param[param_size-1].n;
	param[i].v = param[param_size-1].v;
	param[i].copyflags = param[param_size-1].copyflags;
      }
      --param_size;
      return 1;
    }
  
  return 0;
}

void global_overrides()
{
  char *v;

  if ((v = get_parameter("quiet", 0))) {
    quiet = atoi(v);
    parameter_changes--;
  }
  if ((v = get_parameter("npops", 0))) {
    npops = atoi(v);
    parameter_changes--;
  }
  if ((v = get_parameter("startpopnum", 0))) {
    startpopnum = atoi(v);
    parameter_changes--;
  }
  if ((v = get_parameter("finishpopnum", 0))) {
    finishpopnum = atoi(v);
    parameter_changes--;
  }
  if ((v = get_parameter("ncycles", 0))) {
    ncycles = atoi(v);
    parameter_changes--;
  }
  if ((v = get_parameter("maxgen", 0))) {
    maxgen = atoi(v);
    parameter_changes--;
  }
  if ((v = get_parameter("ckptfreq", 0))) {
    ckptfreq = atoi(v);
    parameter_changes--;
  }
}

void get_pop_overrides()
{
  char *v;
  char num[3];
  static int gpc_sav, ps_sav, cf_sav;
  static float pc_sav, pm_sav, pi_sav;
  static float irf_sav, beta_sav, b_s_tilde_sav, rand_seed_sav;

  static int firstcall = 1;

  if (firstcall) {
    firstcall = 0;
    gpc_sav = genspercycle;
    ps_sav = popsize;
    pc_sav = pcross;
    pm_sav = pmutation;
    pi_sav = pinversion;
    cf_sav = crowding_factor;
    irf_sav = incest_reduction;
    beta_sav = beta;
    b_s_tilde_sav = beta_s_tilde;
  }
  else {
    genspercycle = gpc_sav;
    popsize = ps_sav;
    pcross = pc_sav;
    pmutation = pm_sav;
    pinversion = pi_sav;
    crowding_factor = cf_sav;
    incest_reduction = irf_sav;
    beta = beta_sav;
    beta_s_tilde = b_s_tilde_sav;
    randomseed = rand_seed_sav;
  }

  sprintf(num, "%02d", popno);

  if ((v = get_parameter("genspercycle", num)) 
      || (v = get_parameter("genspercycle", 0)))
    genspercycle = atoi(v);
  if ((v = get_parameter("popsize", num)) 
      || (v = get_parameter("popsize", 0)))
    popsize = atoi(v);
  if ((v = get_parameter("pcross", num))
      || (v = get_parameter("pcross", 0)))
    pcross = atof(v);
  if ((v = get_parameter("pmutation", num))
      || (v = get_parameter("pmutation", 0)))
    pmutation = atof(v);
  if ((v = get_parameter("pinversion", num))
      || (v = get_parameter("pinversion", 0)))
    pinversion = atof(v);
  if ((v = get_parameter("crowding_factor", num))
      || (v = get_parameter("crowding_factor", 0)))
    crowding_factor = atoi(v);
  if ((v = get_parameter("incest_reduction", num))
      || (v = get_parameter("incest_reduction", 0)))
    incest_reduction = atoi(v);
  if ((v = get_parameter("beta", num))
      || (v = get_parameter("beta", 0)))
    beta = atof(v);
  if ((v = get_parameter("beta_s_tilde", num))
      || (v = get_parameter("beta_s_tilde", 0)))
    beta_s_tilde = atof(v);
  if ((v = get_parameter("randomseed", num))
      || (v = get_parameter("randomseed", 0)))
    randomseed = atof(v);

}


/* Get file naming inputs and all other info common to all subpopulations */
/* and set up locking (if needed) on checkpoint and restart file names.   */
/* Called when starting or extending a run. */
void
GetUniversalInputs()
{
  char c, dummy[11];

  if (interactive) {
    printf("     ---------------------*** Data Entry ");
    printf("***---------------------\n\n");
  }

  /* Initialize problem-specific input parameters to INVALID values */
  numfields = 7777;
  ckptfreq = -1;
  local_bestfit.init_fitness = -1.0; /* any non-negative fitness will replace */

  if (many) {
    if (interactive) {
      printf("Number of subpopulations to be simulated, ");
      printf("including all processes -> ");
    }
    ffscanf("npops", infp, "%d", &npops);
    if (interactive) {
      while (npops < 1 || npops > 98) {
	printf("%d is an illegal value for total number of subpopulations.\n",
	       npops);
	printf("Number of subpopulations to be simulated -> ");
	fflush(stdout);
	ffscanf("npops", infp, "%d", &npops);
      }
    }
    
    /* Determine which subpopulations THIS processor is to simulate */
    RunWhichPops();
    
    if (interactive) {
      ncycles = 0;
      while (ncycles < 1) {
	printf("Number of cycles for each subpopulation (no default) -> ");
	ffscanf("ncycles", infp, "%d", &ncycles);
      }
    }
    else
    ffscanf("ncycles", infp, "%d", &ncycles);
  } 

  quiet = 0;
  if (interactive) {
    printf("\nOutput level (quiet) coding:\n 0 = show all;\n");
    printf(" 1 = performance, best of generations;\n 2 = new bests;\n");
    printf(" 3 = no output\nQuiet level desired (%d) -> ", quiet);
  }
  ffscanf("quiet", infp, "%d", &quiet);

  if (!many) {
    if (ckptfreq < 0)
      ckptfreq = 10;  /* set default value */
    if (interactive) {
      printf("Number of generations between writing ");
      printf("of checkpoint files. 0 = only at end of run. (%d) - > ",
	      ckptfreq);
    }
    ffscanf("ckptfreq", infp, "%d", &ckptfreq);
  }

  if (interactive) {
    do {
      printf("\nPrefix for new checkpoint files, must be 6 or fewer");
      printf(" characters\n");
      printf(" (default \"sgackp\" or restart file prefix) -> ");
      if (ffscanf("checkptfileprefix", infp, "%8s", checkptfileprefix) == SKIP){
	/* filename not given, set it empty */
	strcpy (checkptfileprefix, "");
      }
    } while ((int)strlen(checkptfileprefix) > 6); 

    if (strlen(checkptfileprefix))
      printf("   Checkpoint files %s%s.ckp.\n", checkptfileprefix
	     , (many ? "nn" : ""));
  }
  else {
    if (!strlen(checkptfileprefix)) /* didn't take it off command line */
      if (ffscanf("checkptfileprefix", infp, "%8s", checkptfileprefix) == SKIP) 
	strcpy (checkptfileprefix, "");
      else  
	checkptfileprefix[6] = '\0';
    else
      ffscanf("checkptfileprefix", infp, "%8s", dummy);
  }

  if (interactive) {
    do {
      printf("Prefix for files to restart from, 6 chars or fewer\n");
      printf(" (default no restart files) -> ");
      if (ffscanf("restartfileprefix", infp, "%8s", restartfileprefix) == SKIP){
	/* filename not given, set it empty */
	strcpy (restartfileprefix, "");
      }
    }  while ((int)strlen(restartfileprefix) > 6); 

    if (strlen(restartfileprefix)) {
      /* Here, must distinguish between a run of Onepop and Manypops, */
      /* since Manypops will add a 2-digit number to the prefix for file */
      /* names. */
      extending = 1;
      printf("   * EXTENDING a run from %s%s.xxx.\n", restartfileprefix
	      , (many ? "nn" : ""));
    }
    else {
      printf("   * Starting from scratch.\n");
      extending = 0;
    }
  }
  else if (extending) 
      /* if extending is already set, we took a prefix from the command line */
      /* so throw away the one in the input file */
      ffscanf("restartfileprefix", infp, "%8s", dummy);
  else {
      /* Don't yet know whether extending or not... find out now. */
      if (ffscanf("restartfileprefix", infp, "%8s", restartfileprefix) == SKIP)
	strcpy (restartfileprefix, "");
      else {
	extending = 1;
	if ((int)strlen(restartfileprefix) > 6) {
	  fprintf(stderr, "\n*** OOPS!  Restartfileprefix is 6");
	  fprintf(stderr, " characters maximum.  Use shorter prefix.\n");
	  exit(-1);
	}
      }
    }

  if (!(strlen(checkptfileprefix)))
    if (extending)
      strcpy(checkptfileprefix, restartfileprefix);
    else
      strcpy(checkptfileprefix, "sgackp");
    
  /* TO DEBUG LOCKING, ETC., running only one process (and one processor), */
  /* uncomment the next line: */
  /* single_process = FALSE;  */

  if (many)
    /* Read .mst file to learn how inter-population communications to be done */
    SetUpMigration();
  else
    masterfileprefix[0] = '@';

  /* Prior to reading ordinary parameters from std input or input file, */
  /* allow user to input those which are problem-specific, (i.e., not GA */
  /* parameters changeable among subpopulations), plus whatever else     */
  /* he/she wants to put in extra, from a callback.  */
  app_read_prob_params();
  
  if (extending) {
    /* Extending a run from saved checkpoint files, don't ask questions below */
    all_subpops_use_same_parameters = FALSE;
    return;
  }

  c = 'n';
  if (interactive)
    printf("Is this a permutation-type problem? (n) -> ");
  ffscanf("permproblem", infp, "%c", &c);

  if (c == 'y' || c == 'Y') {
    permproblem = 1;
    pinversion = 0.0;
  }
  else {
    permproblem = 0;
  }

  if (!many)
    return;

  /* requests for input about field size, etc, moved to subpop_data() LTWH */

  c = 'y';
  if (interactive) {
    printf("Shall the same parameters be used for ALL subpopulations");
    printf(" (%d through %d)? (y) -> ", startpopnum, finishpopnum);
    fflush(stdout);
  }

  ffscanf("all_subpops_use_same_parameters", infp, "%c", &c);
  if (c == 'n' || c == 'N') {
    all_subpops_use_same_parameters = FALSE;
  }
}

/* Here we read in the text file specifying neighbors of each subpopulation */
/* and communication strategy to be used, etc. */
void
SetUpMigration()
{
  if (interactive) {
    do {
      printf("Prefix for master file, 8 chars or fewer");
      printf(" (default master) -> ");
      if (ffscanf("masterfileprefix", infp, "%10s", masterfileprefix) == SKIP) {
	/* filename not given, set it to "master" */
	strcpy (masterfileprefix, "master");
      }
    } while ((int)strlen(masterfileprefix) > 8);
  }
  
  else {
    char dummy[11];

    if (!strlen(masterfileprefix)) {
      if (ffscanf("masterfileprefix", infp, "%10s", masterfileprefix) == SKIP) 
	/* filename not given, set it to "master" */
	strcpy (masterfileprefix, "master");
      else if ((int)strlen(masterfileprefix) > 8) {
	fprintf(stderr, "\n*** OOPS!  Masterfileprefix is 8");
	fprintf(stderr, " characters maximum.  Use shorter prefix.\n");
	exit(-1);
      }
    }
    else
      ffscanf("masterfileprefix", infp, "%10s", dummy);
  }

  strcpy(masterfilename, masterfileprefix);
  strcat(masterfilename, ".mst");

  if (ReadNeighborTable(masterfilename) == FALSE) {
    masterfileprefix[0] = '@';
    printf("Read of Master file failed.  Assuming isolated subpops.\n");
    fprintf(stderr,"Read of Master file failed.  Assuming isolated subpops.\n");
  }
}

/* Find out if this processor is to run ALL or what subset of the */
/* total set of subpopulations described in the master file read. */
void
RunWhichPops()
{
  startpopnum = 0;
  finishpopnum = npops-1;
  if (interactive) {
    printf("Starting subpopulation number, %d through %d,", 0, npops-1);
    printf(" for this process (default 0) -> ");
  }
  ffscanf("startpopnum", infp, "%d", &startpopnum);

  if (interactive) {
    printf("Last subpopulation");
    printf(" for this process (%d) -> ", finishpopnum);
  }
  else
    finishpopnum = -1;
  ffscanf("finishpopnum", infp, "%d", &finishpopnum);

  if (interactive) {
    while (finishpopnum < startpopnum || npops <= finishpopnum) {
      printf(" %d <= last subpopulation <= %d -> ", 
	      startpopnum, npops-1);
      ffscanf("finishpopnum", infp, "%d", &finishpopnum);
    }
  }
  if (finishpopnum < 0)
    /* Accepted the default npops-1 */
    finishpopnum = npops - 1;
  if (interactive) {
    printf("   * This process will be responsible for subpopulations");
    printf(" %d through %d.\n", startpopnum, finishpopnum);
  }

  /* Now determine whether or not this problem is being run as a SINGLE */
  /* process, or whether some of the subpopulations in this master file */
  /* will be run by other processes and/or processors. */
  if (npops == (finishpopnum - startpopnum + 1))
    single_process = TRUE;
  else
    single_process = FALSE;
}

void subpop_data(must_recalc_best)
BOOL *must_recalc_best;
{
  void problem_inputs();

  problem_inputs();
  pop_inputs(must_recalc_best);
}

void problem_inputs()
{
  char c;
  int k, i, fldlen, fldsiz, prev_end;
  /* Temporary array for calc of std_field_ends. */

  /* initializing from scratch or for first time, must get these values */
  if (!permproblem) {
    if (interactive) {
      printf("Alphabet size is the number of distinct");
      printf(" alleles/locus for your representation\n");
      printf(" (Crossover occurs only at locus boundaries, mutation to");
      printf(" any symbol in alphabet)\n");
      printf("Note:  alpha_size < 2 means sizes differ, get them from\n"); 
      printf(" array field_sizes in app_set_field_sizes().\n"); 
      printf(" Alpha_size (default %d", alpha_size);
      if (alpha_size == 2)
	printf(",  a bit-string chromosome, crossover anywhere) -> ");
      else
	printf(") ->");
      fflush(stdout);
    }

    ffscanf("alpha_size", infp, "%d", &alpha_size);
    if (alpha_size < 2) {
      fprintf(outfp, " \nEach field will have its own number of symbols, as ");
      fprintf(outfp," set in array field_sizes\n in app_set_field_sizes()\n");
    }
  }
  
  /* For any problem type: */
  /* There is an upper limit on numfields for perm problems, where they */
  /* use fixed array storage, but NOT for non-perm problems. */
  if (interactive)
    if (permproblem == 1) {
      printf("Total number of fields to be used,");
      printf(" range 2-%d ", MAXNUMFIELDS);
      if (numfields < 2 || numfields == 7777)
	printf("(no default) -> ");
      else
	printf("(%d) -> ", numfields);
    }
    else {
      printf("Total number of fields to be used, must be >= 2 ");
      if (numfields < 2 || numfields == 7777)
	printf("(no default) -> ");
      else
	printf("(%d) -> ", numfields);
    }
  ffscanf("numfields", infp, "%d", &numfields);

  while (numfields == 7777 || numfields < 2
	 || (permproblem == 1 && numfields > MAXNUMFIELDS)) {
    if (interactive)
      if (permproblem == 1) {
	printf("Total number of fields to be used,");
	printf(" range 2-%d, (no default) -> ", MAXNUMFIELDS);
      }
      else {
	printf("Total number of fields to be used, must be >= 2,");
	printf(" (no default) -> ");
      }
    ffscanf("numfields", infp, "%d", &numfields);
  }
  
  if (permproblem == 1) {
    /* Now, need following if a perm problem, else no effect */
    alpha_size = numfields;
  }

  if (!permproblem && alpha_size == 2) {
    if (interactive)
      printf("Use superuniform initialization? (%c) -> ",
	      superuniform_initialization == TRUE ? 'y' : 'n');
    ffscanf("superuniform", infp, "%c", &c);
    if (c == 'y' || c == 'Y') {
      superuniform_initialization = TRUE;
    }
  }
}

void pop_inputs(must_recalc_best)
BOOL *must_recalc_best;
{
  char c, answer;
  int oldpopsize, k;

  if (many) {
    if (interactive) {
      printf("Number of generations per cycle ");
      if (genspercycle < 1)
	printf("(no default) -> ");
      else
	printf("(%d) -> ", genspercycle);
      fflush(stdout);
    }
    ffscanf("genspercycle", infp, "%d", &genspercycle);
    if (interactive) 
      while (genspercycle < 1) {
	printf("Number of generations per cycle (no default) -> ");
	ffscanf("genspercycle", infp, "%d", &genspercycle);
      }
  }  

  oldpopsize = popsize;
  if (interactive) 
    if (popsize < 1)
      printf("Population size, must be even number (no default) -> ");
    else
      printf("Population size, must be even number (%d) -> ", popsize);
  ffscanf("popsize", infp, "%d", &popsize);
  if (interactive) 
    while (popsize < 2 || (popsize%2)) {
      printf("Population size, must be even number (no default) -> ");
      ffscanf("popsize", infp, "%d", &popsize);
    }
  else if ((popsize % 2)) {
    printf("Sorry! only even population sizes are allowed. \n"); 
    printf("Incrementing popsize by one.\n");
    popsize++;
  }
  else if (popsize < 2) {
    printf("Sorry! minimum population size is 2.\n");
    popsize = 2;
  }
  
  if (extending) 
    /* If REDUCING population size, need to be sure that the string indicated */
    /* by variable "bestnow" is still in range of the population size.  If    */
    /* using elitism, must copy best guy INTO new range, replacing a random   */
    /* individual.  If NOT using elitism, just recalculate the new best guy   */
    /* after the population is initialized, but BEFORE migration, which may   */
    /* try to USE the best guy for incest_reduction. */
    if (popsize < oldpopsize && popsize < bestnow) 
      *must_recalc_best = TRUE; 

  if (interactive)
    printf("Print chromosome strings? (%c) -> ", 
	    (printstrings ? 'y' : 'n'));
  ffscanf("printstrings", infp, "%c", &answer);
  if (answer == 'n' || answer == 'N')
    printstrings = 0;
  else if (answer == 'y' || answer == 'Y')
    printstrings = 1;
  
  if (interactive) {
    printf("Crossover probability, 0 <= pcross <= 1.0 ");
    if (pcross < 0.)
      pcross = 0.7;
    printf("(%.4f) -> ", pcross);
  }
  ffscanf("pcross", infp, "%f", &pcross);
  if (interactive) {
    while (pcross < 0. || 1.0 < pcross) {
      printf("Crossover probability, 0 <= pcross <= 1.0 (no default) -> ");
      ffscanf("pcross", infp, "%f", &pcross);
    }
  }

  if (interactive) {
    printf("Mutation probability%s, 0 <= pmutation <= 1.0 ", 
	    permproblem ? "/chromosome" : "/field");
    if (pmutation < 0.)
      if(permproblem)
	pmutation = 0.2;
      else
	pmutation = 0.2/numfields;   
    printf("(%.6f) -> ", pmutation);
  }
  ffscanf("pmutation", infp, "%f", &pmutation);
  if (interactive) 
    while (pmutation < 0. || 1.0 < pmutation) {
      printf("Mutation probability, 0 <= pmutation <= 1.0 (no default) -> ");
      ffscanf("pmutation", infp, "%f", &pmutation);
    }

  if (!permproblem) {
    if (interactive) {
      if (pinversion < 0)
	pinversion = 0.0;
      else if (1.0 < pinversion)
	pinversion = 1.0;

      printf("Inversion probability/generation,");
      printf(" 0 <= pinversion <= 1.0 (%.4f) -> ", pinversion);
    }
    ffscanf("pinversion", infp, "%f", &pinversion);
    if (interactive) 
      while (pinversion < 0 || 1.0 < pinversion) {
	if (pinversion < 0)
	  pinversion = 0.0;
	else if (1.0 < pinversion)
	  pinversion = 1.0;
	
	printf("Inversion probability/generation.");
	printf(" 0 <= pinversion <= 1.0 (%.4f) -> ", pinversion);
	ffscanf("pinversion", infp, "%f", &pinversion);
      }
  }
  
  /* Process input for "Autocontrolled selection" */
  autocontrol_selection = 1;
  if (interactive)
    printf("Want to use automatic control of selection with Boltzmann scaling?(%c) -> ", 
	    (autocontrol_selection ? 'y' : 'n'));
  answer = 'y';
  ffscanf("autocontrol_selection", infp, "%c", &answer);
  if (answer == 'n' || answer == 'N') {
      autocontrol_selection = 0;
      beta = -1.;
      if (interactive) {
	printf("Beta for Boltzmann-scaled fitnesses, -1. (OFF)"); 
	printf(" or 0. <= beta <= 5.0 ");
	if (beta < 0.)
	  printf("(OFF) -> ");
	else
	  printf("(%.4f) -> ", beta);
      }
      ffscanf("beta", infp, "%f", &beta);
      if (interactive) {
	while ((beta < 0. && abs( beta + 1.) > 0.01) || 5.0 < beta ) {
	  printf("Beta for Boltzmann-scaled fitnesses, -1. (OFF)"); 
	  printf(" or 0. <= beta <= 5.0 (%.4f) -> ", beta);
	  ffscanf("beta", infp, "%f", &beta);
	}
      }
  }
  else if (answer == 'y' || answer == 'Y') {
      autocontrol_selection = 1;
      /* Since I agree with Shapiro that automatic selection should use */
      /* Stochastic Universal Sampling, I require that user load suselect.c */
      /* if this automatic option is chosen.  Checked here. */
      if (which_selection[1] != 'u') {
	fprintf (stderr, "OOPS, Automatic control of selection uses ");
	fprintf (stderr, "suselect.c, Stochastic Universal Sampling,\n");
	fprintf (stderr, "and that is not the selection method loaded.\n");
	fprintf (stderr, "Please load that and rerun.");
	exit (-1);
      }
      /* Value 0.5 below (beta0) was chosen from Shapiro's graphs (called */
      /* beta_zero or beta_s in various papers from their group) as a */
      /* point where decrease of variance was small while still allowing */
      /* some increase in mean fitness.  It might be usefully varied */
      /* from as little as 0.1 to as much as 0.6 depending on the desired */
      /* length of run for the difficulty of the problem, and the effects */
      /* of the crossover and mutation operations on increasing diversity */
      /* of fitness of the population.  */

      beta_s_tilde = 0.5 * sqrt(2.0 * log((double)popsize));
      if (interactive) {
	printf("\nBeta will be calculated each generation, by dividing ");
	printf("beta_s_tilde");
	printf("\nby the standard deviation of the current fitness ");
	printf("distribution.");
	printf("\nBeta_s_tilde will default to a \"reasonable\" value, the");
	printf("\nsquare root of the log of the population (subpop) size \n");
	printf("times a value of 0.5 selected from Figure 3 of Shapiro, J. \n");
	printf("and Pruegel-Bennett, A., Maximum Entropy Analysis of \n");
	printf("Genetic Algorithm Operators, in Fogarty, T.C. (ed.), \n");
	printf("Evolutionary Computing, Proc. of AISB Workshop, Sheffield,\n");
	printf(" UK, Springer, New York, Lecture Notes in Computer Science,\n");
	printf(" April, 1995, unless you override it here (feel free):");
	printf("\nBeta_s_tilde for Boltzmann-scaled fitnesses, ");
	printf(">= 0. (%.4f) ->", beta_s_tilde);
      }
      ffscanf("beta_s_tilde", infp, "%f", &beta_s_tilde);
      if (interactive) {
	while (beta_s_tilde < 0. ) {
	  printf("Beta_s_tilde for Boltzmann-scaled fitnesses >=");
	  printf("0. (%d) -> ", beta_s_tilde);
	  ffscanf("beta_s_tilde", infp, "%f", &beta_s_tilde);
	}
      }
  }

  if (!autocontrol_selection && beta < 0.) {
    /* Not using Boltzmann scaling or autocontrol_selection */
    if (interactive) {
      printf("Note: sigma truncation or linear scaling require");
      printf(" window scaling to be off (= -1).\n");
      printf("Length (# generations) of scaling window (default");
      if (scaling_window == -1)
	printf(" off) -> ");
      else
	printf(" %d) -> ", scaling_window);
    }
    ffscanf("scaling_window", infp, "%d", &scaling_window);
    if (scaling_window >= MAXSCALINGWINDOW) {
      printf(" %d too large. Scaling window set to its maximum, %d.\n",
	      scaling_window, MAXSCALINGWINDOW - 1);
      scaling_window = MAXSCALINGWINDOW - 1;
    }

    if (scaling_window == -1) {
      if (interactive) {
	printf(
		"Coefficient for sigma truncation, 0 == off (default off) -> ");
	fflush(stdout);
      }
      ffscanf("sigma_trunc", infp, "%f", &sigma_trunc);

      if (interactive) {
	printf("Linear scaling multiplier, 0 = scaling off.\n");
	printf(" Also used as best/mean if");
	printf(" ranking is the basis for selection.\n Scaling ");
	printf(" multipler, best/mean, 0 or > 1.0 (%.3f) -> ",
		scalemult);
      }
      ffscanf("scalemult", infp, "%f", &scalemult);

      while (scalemult && scalemult <= 1.0) {
	  fprintf(stderr, " %f is illegal value, must be 0 or > 1.0 -> ", 
		  scalemult);
	  ffscanf("scalemult", infp, "%f", &scalemult);
	}
    }
  }
  if (interactive) 
    printf("Crowding factor integer, 0 = no crowding (%d) -> ",
	    crowding_factor);
  ffscanf("crowding_factor", infp, "%d", &crowding_factor);

  if (crowding_factor > 0) {
    if (interactive) 
      printf("Incest reduction, 0 = off (%d) -> ", incest_reduction);
    ffscanf("incest_reduction", infp, "%d", &incest_reduction);
  }

  if (interactive) {
    printf("Lower limit for 'good' individual,");
    printf(" in std. deviations above mean,\n for convergence ");
    printf("calculation (7.0 = no lower limit) (%.1f) -> ",
	    conv_sigma_coeff);
  }
  ffscanf("conv_sigma_coeff", infp, "%f", &conv_sigma_coeff);


  if (interactive) {
    printf("Generations between convergence printouts (%d) -> ",
	    convinterval);
  }
  ffscanf("convinterval", infp, "%d", &convinterval);

  /* any application-dependent global input */
  if (extending)
    app_data(TRUE);             /* flag says REVISING */
  else
    app_data(FALSE);            /* flag says INITIALIZING */

  /* If using tournament selection, desired tourney size  needed. */
  /* Moved here from tselect.c in Release 2.37 (April 4, 1995) */
  if (which_selection[0] == 't') {
    if (interactive)
      printf("Tournament size for selection (%d) -> ", 
	      tourneysize);
    ffscanf("tourneysize", infp, "%d", &tourneysize);
    if (interactive) 
      while (tourneysize > popsize) {
	printf(" Tournament size, %d, cannot exceed population size, %d -> ",
	       tourneysize, popsize);
	ffscanf("tourneysize", infp, "%d", &tourneysize);
      } 
    else if (tourneysize > popsize) {
      printf("FATAL: Tournament size (%d) -> popsize (%d)\n",
	      tourneysize, popsize);
      exit(-1);
    }
  }
}

void usage()
{
  fprintf(stderr, "Usage: %s [-ioce%s <file name|prefix>]",
	  (many ? "Manypops": "Onepop"), (many ? "m": ""));
  fprintf(stderr, " [-p <tag>=<value>] ...\n\n");
  fprintf(stderr, "  File names:\n");
  fprintf(stderr, "    -i <file name>    file of input parameter values");
  fprintf(stderr, " (omitted==>interactive)\n");
  fprintf(stderr, "    -o <file name>    file for run time output\n");
  fprintf(stderr, "    -c <file prefix>");
  fprintf(stderr, "  checkpointfile prefix (6 char max)\n");
  fprintf(stderr, "    -e <file prefix>");
  fprintf(stderr, "  restartfile prefix for extending a run (6 char max)\n");
  if (many) {
    fprintf(stderr, "    -m <file prefix>  masterfile prefix");
    fprintf(stderr, " for migration patterns (8 char max)\n");
  } 
  fprintf(stderr, "\n  Global parameters:\n");
  fprintf(stderr, "    -p quiet=<output level, 0-3>\n");
  if (many) {
    fprintf(stderr, "    -p npops=<# of populations to be run>\n");
    fprintf(stderr, "    -p startpopnum=<first population for this");
    fprintf(stderr, " process/processor, 0 to npops-1>\n");
    fprintf(stderr, "    -p finishpopnum=<last population for this");
    fprintf(stderr, " process/processor, 0 to npops-1>\n");
    fprintf(stderr, "    -p ncycles=<# of cycles to be run>\n");
  }
  else {
    fprintf(stderr, "    -p maxgen=<# of generations to be run>\n");
    fprintf(stderr, "    -p ckptfreq=<# of generations between checkpoints>\n");
  }
  fprintf(stderr, "\n  Population specific parameters:\n");
  fprintf(stderr, "    -p pcross=<%% probability of crossover>\n");
  fprintf(stderr, "    -p pmutation=<%% probability of mutation>\n");
  fprintf(stderr, "    -p pinversion=<%% probability of inversion>\n");
  fprintf(stderr, "    -p crowding_factor=<crowding factor>\n");
  fprintf(stderr, "    -p incest_reduction=<incest reduction factor>\n");
  fprintf(stderr, "    -p beta=<Beta for manual Boltzmann determ. sel'n>\n");
  fprintf(stderr, "    -p beta_s_tilde=<Base value beta_s_tilde from which ");
  fprintf(stderr, "beta \n          is calculated in automatic Boltzmann ");
  fprintf(stderr, "selection>\n");
  fprintf(stderr, "    -p popsize=<# of individuals/population>\n");
  if (many) {
    fprintf(stderr, "    -p genspercycle=<# of generations per cycle>\n");
    fprintf(stderr, "\n    Add a 2 digit population number to tag to change");
    fprintf(stderr, " the value of a \n    population-specific parameter for");
    fprintf(stderr, " only a single population, i.e.,\n");
    fprintf(stderr, "    -p pcross02=0.05  will only override for pop #2\n");
  }
  fprintf(stderr, "\n");
}

/*
void write_parameter_database(FILE *f)
{
  int i, j;
     
  fwrite(&param_size, sizeof(int), 1, f);
  for (i = 0; i < param_size; ++i)
    {
      j = strlen(param[i].n);
      fwrite(&j, sizeof(int), 1, f);
      fwrite(param[i].n, j, 1, f);

      j = strlen(param[i].v);
      fwrite(&j, sizeof(int), 1, f);
      fwrite(param[i].v, j, 1, f);
    }
}

void read_parameter_file(char *filename)
{
  FILE *f;
  char buffer[MAXPARAMLINELENGTH+1];
  int line = 0;
  int errors = 0;
  
  if (!(f = fopen(filename, "r"))) {
    fprintf(stderr, "Can't open parameter file \"%s\".", filename);
    exit(1);
  }
  
  while (fgets(buffer, MAXPARAMLINELENGTH, f)) {
    ++line;

    if (!parse_one_parameter(buffer)) {
      errors++;
      fprintf(stderr, "%s line %d: syntax error.", filename, line);
    }
  }
  
  fclose(f);
  
  if (errors)
    exit(1);
}

void binary_parameter(char *name, int value)
{
  char *param = get_parameter(name, 0);
  char string[2];
  char *i;
  
  if (param != NULL) {
    for (i = param; *i; ++i)
      *i = tolower(*i);
	  
    if (strcmp(param, "true") == 0 ||
	strcmp(param, "t") == 0 ||
	strcmp(param, "on") == 0 ||
	strcmp(param, "yes") == 0 ||
	strcmp(param, "y") == 0 ||
	strcmp(param, "1") == 0)
      value = 1;
    else if (strcmp(param, "false") == 0 ||
	     strcmp(param, "f") == 0 ||
	     strcmp(param, "off") == 0 ||
	     strcmp(param, "no") == 0 ||
	     strcmp(param, "n") == 0 ||
	     strcmp(param, "0") == 0)
      value = 0;
    else
      fprintf(stderr,
	     "\"%s\" is not a legal value for \"%s\"; override not accepted.\n",
	     param, name);
  }
  
  sprintf(string, "%d", !!value);
  add_parameter(name, string, PARAM_COPY_VALUE|PARAM_COPY_NAME);
}
*/
